import asyncio
from datetime import datetime

from PyQt5.QtWidgets import QFrame
from PyQt5.QtWidgets import QPushButton
from PyQt5.QtWidgets import QWidget
from PyQt5.QtWidgets import QToolButton
from PyQt5.QtWidgets import QLineEdit
from PyQt5.QtWidgets import QMenu
from PyQt5.QtWidgets import QSpacerItem
from PyQt5.QtWidgets import QAction
from PyQt5.QtWidgets import QVBoxLayout
from PyQt5.QtWidgets import QWidgetAction
from PyQt5.QtWidgets import QLabel
from PyQt5.QtWidgets import QApplication
from PyQt5.QtWidgets import QComboBox
from PyQt5.QtWidgets import QListView
from PyQt5.QtWidgets import QTextEdit
from PyQt5.QtWidgets import QHBoxLayout
from PyQt5.QtWidgets import QSizePolicy
from PyQt5.QtWidgets import QTextBrowser
from PyQt5.QtWidgets import QToolTip

from PyQt5.QtWebEngineWidgets import QWebEngineView
from PyQt5.QtWebEngineWidgets import QWebEnginePage

from PyQt5.QtCore import QRect
from PyQt5.QtCore import QEvent
from PyQt5.QtCore import QRegExp
from PyQt5.QtCore import QTimer
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtCore import Qt
from PyQt5.QtCore import QUrl
from PyQt5.QtCore import QMimeData
from PyQt5.QtCore import QSize
from PyQt5.QtCore import QFile
from PyQt5.QtCore import QIODevice
from PyQt5.QtCore import QVariant
from PyQt5.QtCore import QPoint

from PyQt5.QtGui import QRegExpValidator
from PyQt5.QtGui import QImage
from PyQt5.QtGui import QPixmap
from PyQt5.QtGui import QTextCursor
from PyQt5.QtGui import QKeySequence

from pygments import highlight
from pygments.formatters import HtmlFormatter
from pygments.lexers import get_lexer_by_name

from galacteek.ipfs.stat import StatInfo
from galacteek.ipfs.wrappers import ipfsOp
from galacteek.ipfs.cidhelpers import IPFSPath
from galacteek import ensure
from galacteek.core import parseDate
from galacteek.core.asynclib import asyncify
from galacteek.dweb.markdown import markitdown

from .helpers import getIcon
from .helpers import getIconFromIpfs
from .helpers import disconnectSig
from .helpers import sizeFormat
from .helpers import messageBox
from .i18n import iCancel
from .i18n import iUnknown
from .i18n import iLocalHashmarksCount
from .i18n import iSearchHashmarks
from .i18n import iSearchHashmarksAllAcross
from .i18n import iSearchUseShiftReturn
from .i18n import iHashmarkInfoToolTip


class GalacteekTab(QWidget):
    def __init__(self, gWindow, **kw):
        super(GalacteekTab, self).__init__(gWindow)
        self.vLayout = QVBoxLayout(self)
        self.setLayout(self.vLayout)

        self.gWindow = gWindow
        self.setAttribute(Qt.WA_DeleteOnClose)

    def tabIndex(self, w=None):
        return self.gWindow.tabWidget.indexOf(w if w else self)

    def setTabName(self, name, widget=None):
        idx = self.tabIndex(w=widget)
        if idx >= 0:
            self.gWindow.tabWidget.setTabText(idx, name)

    def addToLayout(self, widget):
        self.vLayout.addWidget(widget)

    def onClose(self):
        return True

    @ipfsOp
    async def initialize(self, op):
        pass

    @property
    def app(self):
        return self.gWindow.app

    @property
    def loop(self):
        return self.app.loop

    @property
    def profile(self):
        return self.app.ipfsCtx.currentProfile


class HorizontalLine(QFrame):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setFrameShape(QFrame.HLine)
        self.setFrameShadow(QFrame.Sunken)
        self.setStyleSheet('''
            QFrame {
                background-color: #4a9ea1;
            }
        ''')


class URLDragAndDropProcessor:
    fileDropped = pyqtSignal(QUrl)
    ipfsObjectDropped = pyqtSignal(IPFSPath)

    def dragEnterEvent(self, event):
        mimeData = event.mimeData()

        if mimeData.hasUrls() or mimeData.hasText():
            event.acceptProposedAction()

    def dropEvent(self, event):
        mimeData = event.mimeData()

        if mimeData is None:
            return

        if mimeData.hasUrls():
            for url in mimeData.urls():
                if not url.isValid():
                    continue

                if url.scheme() == 'file' and url.isValid():
                    self.fileDropped.emit(url)
                else:
                    path = IPFSPath(url.toString(), autoCidConv=True)
                    if path.valid:
                        self.ipfsObjectDropped.emit(path)
        elif mimeData.hasText():
            text = mimeData.text()
            path = IPFSPath(text, autoCidConv=True)

            if path.valid:
                self.ipfsObjectDropped.emit(path)

        event.acceptProposedAction()


class CheckableToolButton(QToolButton):
    def __init__(self, toggled=None, icon=None, parent=None):
        super(CheckableToolButton, self).__init__(parent)
        self.setCheckable(True)
        self.setAutoRaise(True)

        if toggled:
            self.toggled.connect(toggled)

        if icon:
            self.setIcon(icon)


class IPFSPathClipboardButton(QToolButton):
    def __init__(self, ipfsPath, parent=None):
        super(IPFSPathClipboardButton, self).__init__(parent)

        self.setEnabled(False)
        self.setIcon(getIcon('clipboard.png'))
        self.app = QApplication.instance()
        self.path = ipfsPath
        self.clicked.connect(self.onClicked)

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, path):
        if isinstance(path, IPFSPath) and path.valid:
            self.setEnabled(True)
            self.setToolTip('Copy to clipboard: {}'.format(str(path)))
            self._path = path

    def onClicked(self):
        if isinstance(self.path, IPFSPath) and self.path.valid:
            self.app.setClipboardText(str(self.path))


class IPFSUrlLabel(QLabel):
    def __init__(self, ipfsPath, invalidPathLabel='Invalid path', parent=None):
        super(IPFSUrlLabel, self).__init__(parent)
        self.app = QApplication.instance()

        self.invalidPathMessage = invalidPathLabel
        self.path = ipfsPath
        self.linkActivated.connect(self.onLinkClicked)

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, val):
        self._path = val

        if self.path and self.path.valid:
            self.url = QUrl(self.path.ipfsUrl)
            self.setText('<a href="{url}">{url}</a>'.format(
                url=self.url.toString()))
        else:
            self.url = None
            self.setText(self.invalidPathMessage)

    def dragEnterEvent(self, event):
        mimeData = event.mimeData()

        if mimeData.hasUrls() or mimeData.hasText():
            event.acceptProposedAction()

    def mimeData(self, indexes):
        mimedata = QMimeData()
        mimedata.setUrls([self.url])
        return mimedata

    def onLinkClicked(self, urlString):
        ensure(self.app.resourceOpener.open(self.path))


class LabelWithURLOpener(QLabel):
    """
    QLabel which opens http/https URLs in a tab.
    Used by the 'About' dialog
    """

    def __init__(self, text, parent=None):
        super(LabelWithURLOpener, self).__init__(text, parent)
        self.app = QApplication.instance()
        self.linkActivated.connect(self.onLinkClicked)

    def onLinkClicked(self, urlString):
        url = QUrl(urlString)
        if url.isValid() and url.scheme() in ['http', 'https']:
            tab = self.app.mainWindow.addBrowserTab()
            tab.enterUrl(url)


class PopupToolButton(QToolButton, URLDragAndDropProcessor):
    def __init__(self, icon=None, parent=None, menu=None,
                 mode=QToolButton.MenuButtonPopup, acceptDrops=False,
                 menuSizeHint=None, toolTipsVisible=True):
        super(PopupToolButton, self).__init__(parent)

        self.menu = menu if menu else QMenu(self)
        self.menu.setToolTipsVisible(toolTipsVisible)
        self.setPopupMode(mode)
        self.setMenu(self.menu)
        self.setAcceptDrops(acceptDrops)

        if icon:
            self.setIcon(icon)


class IPFSObjectToolButton(QToolButton):
    """
    Used in the quickaccess toolbar
    """
    deleteRequest = pyqtSignal()

    def __init__(self, ipfsPath, icon=None, parent=None):
        super(IPFSObjectToolButton, self).__init__(parent=parent)
        if icon:
            self.setIcon(icon)

    def mousePressEvent(self, event):
        button = event.button()

        if button == Qt.RightButton:
            menu = QMenu(self)
            menu.addAction('Remove', lambda: self.deleteRequest.emit())
            menu.exec(self.mapToGlobal(event.pos()))

        super().mousePressEvent(event)


class HashmarkToolButton(QToolButton):
    """
    Used in the quickaccess toolbar
    """
    deleteRequest = pyqtSignal()

    def __init__(self, mark, icon=None, parent=None):
        super(HashmarkToolButton, self).__init__(parent=parent)
        if icon:
            self.setIcon(icon)

        self._hashmark = mark

    @property
    def hashmark(self):
        return self._hashmark

    def mousePressEvent(self, event):
        button = event.button()

        if button == Qt.RightButton:
            menu = QMenu(self)
            menu.addAction('Remove', lambda: self.deleteRequest.emit())
            menu.exec(self.mapToGlobal(event.pos()))

        super().mousePressEvent(event)


class _HashmarksCommon:
    def makeAction(self, path, mark, loadIcon=True):
        tLenMax = 64
        title = mark['metadata'].get('title')
        description = mark['metadata'].get('description')
        datecreated = mark.get('datecreated')
        icon = mark.get('icon')

        if title and len(title) > tLenMax:
            title = '{0} ...'.format(title[0:tLenMax])

        dt = parseDate(datecreated)
        if dt:
            dateHuman = dt.isoformat(' ', timespec='minutes')
        else:
            dateHuman = iUnknown()

        action = QAction(title, self)
        action.setToolTip(
            iHashmarkInfoToolTip(path, title, description, dateHuman))

        action.setData({
            'path': path,
            'mark': mark,
            'iconcid': icon
        })

        if not icon or not loadIcon:
            # Default icon
            action.setIcon(getIcon('ipfs-logo-128-white-outline.png'))
        else:
            ensure(self.loadMarkIcon(action, icon))

        return action

    @ipfsOp
    async def loadMarkIcon(self, ipfsop, action, iconCid):
        data = action.data()

        if isinstance(data, dict) and 'iconloaded' in data:
            return

        icon = await getIconFromIpfs(ipfsop, iconCid)

        if icon:
            await ipfsop.sleep()

            action.setIcon(icon)
            data['iconloaded'] = True
            action.setData(data)


class HashmarkMgrButton(PopupToolButton, _HashmarksCommon):
    hashmarkClicked = pyqtSignal(str, str)

    def __init__(self, marks, iconFile='hashmarks.png',
                 maxItemsPerCategory=128, parent=None):
        super(HashmarkMgrButton, self).__init__(parent=parent,
                                                mode=QToolButton.InstantPopup)

        self.setObjectName('hashmarksMgrButton')
        self.menu.setObjectName('hashmarksMgrMenu')
        self.menu.setToolTipsVisible(True)
        self.hCount = 0
        self.marks = marks
        self.cMenus = {}
        self.maxItemsPerCategory = maxItemsPerCategory
        self.setIcon(getIcon(iconFile))

        disconnectSig(self.marks.changed, self.onChanged)

        self.marks.changed.connect(self.onChanged)
        self.marks.markAdded.connect(self.onMarkAdded)
        self.marks.markDeleted.connect(self.onMarkDeleted)

    def onMarkAdded(self):
        ensure(self.updateIcons())

    def onMarkDeleted(self, category, path):
        if category in self.cMenus:
            menu = self.cMenus[category]

            for action in menu.actions():
                data = action.data()
                if data and data['path'] == path:
                    menu.removeAction(action)

            if menu.isEmpty():
                menu.hide()

    def onChanged(self):
        self.updateMenu()

    async def updateIcons(self):
        for mName, menu in self.cMenus.items():
            await asyncio.sleep(0)

            for action in menu.actions():
                data = action.data()

                if 'iconloaded' not in data and data['iconcid']:
                    ensure(self.loadMarkIcon(action, data['iconcid']))

    def updateMenu(self):
        self.hCount = 0
        categories = self.marks.getCategories()

        for category in categories:
            marks = self.marks.getCategoryMarks(category)
            mItems = marks.items()

            if len(mItems) not in range(1, self.maxItemsPerCategory):
                continue

            if category not in self.cMenus:
                self.cMenus[category] = QMenu(category)
                self.cMenus[category].setToolTipsVisible(True)
                self.cMenus[category].triggered.connect(self.linkActivated)
                self.cMenus[category].setObjectName('hashmarksMgrMenu')
                self.menu.addMenu(self.cMenus[category])

            menu = self.cMenus[category]
            menu.setIcon(getIcon('stroke-cube.png'))

            def exists(path):
                for action in menu.actions():
                    if action.data()['path'] == path:
                        return action

            if len(mItems) in range(1, self.maxItemsPerCategory):
                for path, mark in mItems:
                    self.hCount += 1

                    if exists(path):
                        continue

                    menu.addAction(self.makeAction(path, mark))
            else:
                menu.hide()

        self.setToolTip(iLocalHashmarksCount(self.hCount))

    def linkActivated(self, action):
        data = action.data()
        path, mark = data['path'], data['mark']

        if mark:
            if 'metadata' not in mark:
                return
            self.hashmarkClicked.emit(
                path,
                mark['metadata']['title']
            )


class HashmarksSearchWidgetAction(QWidgetAction):
    pass


class HashmarksSearchLine(QLineEdit):
    """
    The hashmarks search line edit.

    Run the search when Shift + Return is pressed.
    """
    searchRequest = pyqtSignal(str)
    returnNoModifier = pyqtSignal()

    def mouseDoubleClickEvent(self, event):
        event.ignore()

    def keyPressEvent(self, event):
        if event.type() == QEvent.KeyPress:
            modifiers = event.modifiers()
            key = event.key()

            if modifiers & Qt.ShiftModifier:
                if key == Qt.Key_Return:
                    searchText = self.text().strip()
                    if searchText:
                        self.searchRequest.emit(searchText)

                    return
            else:
                if event.key() == Qt.Key_Return:
                    self.returnNoModifier.emit()
                    return

        return super().keyPressEvent(event)


class HashmarksSearcher(PopupToolButton, _HashmarksCommon):
    hashmarkClicked = pyqtSignal(str, str)

    def __init__(self, iconFile='hashmarks-library.png', parent=None,
                 mode=QToolButton.InstantPopup):
        super(HashmarksSearcher, self).__init__(parent=parent, mode=mode)

        self.app = QApplication.instance()

        self._catalog = {}
        self._sMenus = []
        self.setObjectName('hashmarksSearcher')

        self.setIcon(getIcon(iconFile))
        self.setShortcut(QKeySequence('Ctrl+Alt+h'))

        self.menu.setTitle(iSearchHashmarks())
        self.menu.setIcon(getIcon(iconFile))
        self.menu.setObjectName('hashmarksSearchMenu')
        self.menu.setToolTipsVisible(True)
        self.menu.aboutToShow.connect(self.aboutToShowMenu)
        self.configureMenu()

    @property
    def searchesCount(self):
        return len(self.menu.actions()) - len(self.protectedActions)

    @property
    def protectedActions(self):
        return [self.searchWAction, self.clearAction]

    def isProtectedAction(self, action):
        for pAction in self.protectedActions:
            if action is pAction:
                return True
        return False

    def aboutToShowMenu(self):
        self.searchLine.setFocus(Qt.OtherFocusReason)

    def configureMenu(self):
        self.clearAction = QAction(
            getIcon('clear-all.png'),
            'Clear searches',
            self.menu,
            triggered=self.onClearSearches
        )
        self.clearAction.setEnabled(False)
        self.searchLine = HashmarksSearchLine(self.menu)
        self.searchLine.setObjectName('hLibrarySearch')

        mediumSize = self.fontMetrics().size(0, '*' * (40))
        self.searchLine.setMinimumSize(QSize(mediumSize.width(), 32))
        self.searchLine.setFocusPolicy(Qt.StrongFocus)

        self.searchLine.returnNoModifier.connect(self.onReturnNoShift)
        self.searchLine.searchRequest.connect(self.onSearch)
        self.searchLine.setToolTip(iSearchHashmarksAllAcross())

        self.searchLine.setValidator(
            QRegExpValidator(QRegExp(r"[\w\s@\+\-_./?'\"!#]+")))
        self.searchLine.setMaxLength(96)
        self.searchLine.setClearButtonEnabled(True)

        self.searchWAction = HashmarksSearchWidgetAction(self.menu)
        self.searchWAction.setDefaultWidget(self.searchLine)
        self.menu.addAction(self.searchWAction)
        self.menu.setDefaultAction(self.searchWAction)
        self.menu.addAction(self.clearAction)

    def onReturnNoShift(self):
        QToolTip.showText(self.searchLine.mapToGlobal(
            QPoint(30, self.searchLine.height())),
            iSearchUseShiftReturn(),
            self.searchLine, QRect(0, 0, 0, 0), 1800)

    def onSearch(self, text):
        for action in self.menu.actions():
            if self.isProtectedAction(action):
                continue
            if action.text() == text:
                self.menu.removeAction(action)

        ensure(self.searchInCatalog(text))

    @asyncify
    async def searchInCatalog(self, text):
        resultsMenu = QMenu(text, self.menu)
        resultsMenu.setToolTipsVisible(True)
        resultsMenu.setObjectName('hashmarksSearchMenu')
        resultsMenu.triggered.connect(
            lambda action: self.linkActivated(
                action, closeMenu=True))
        resultsMenu.menuAction().setData({
            'query': text,
            'datecreated': datetime.now()
        })

        sources = [marks for sender, marks in self._catalog.items()]
        sources.append(self.app.marksLocal)
        showMax = 128
        added = 0

        for marks in sources:
            await asyncio.sleep(0)

            async for hashmark in marks.searchAllByMetadata({
                'path': text,
                'title': text,
                'description': text,
                'comment': text
            }):

                await asyncio.sleep(0)
                resultsMenu.addAction(self.makeAction(
                    hashmark.path, hashmark.markData))

                if added > showMax:
                    break

                added += 1

        if len(resultsMenu.actions()) > 0:
            self.menu.addMenu(resultsMenu)
            resultsMenu.setIcon(getIcon('search-engine.png'))

        self.clearAction.setEnabled(self.searchesCount > 0)
        self.searchLine.clear()

    def onClearSearches(self):
        for action in self.menu.actions():
            if not self.isProtectedAction(action):
                self.menu.removeAction(action)

        self.clearAction.setEnabled(self.searchesCount > 0)

    def register(self, nodeId, ipfsMarks):
        if nodeId in self._catalog.keys():
            del self._catalog[nodeId]

        self._catalog[nodeId] = ipfsMarks

    def linkActivated(self, action, closeMenu=False):
        data = action.data()
        path, mark = data['path'], data['mark']

        if mark:
            if 'metadata' not in mark:
                return

            self.hashmarkClicked.emit(
                path,
                mark['metadata']['title']
            )

            self.searchLine.clear()
            if closeMenu:
                self.menu.hide()


class ImageWidget(QLabel):
    """
    Displays an image stored in IPFS into a QLabel
    """

    def __init__(self, scaleWidth=64, parent=None):
        super(ImageWidget, self).__init__(parent)

        self._imgPath = None
        self._scaleWidth = scaleWidth

    @property
    def imgPath(self):
        return self._imgPath

    @ipfsOp
    async def load(self, ipfsop, path):
        try:
            imgData = await ipfsop.waitFor(
                ipfsop.client.cat(path), 8
            )

            if not imgData:
                raise Exception('Failed to load image')

            img1 = QImage()
            img1.loadFromData(imgData)
            img = img1.scaledToWidth(self._scaleWidth)

            self.setPixmap(QPixmap.fromImage(img))
            self._imgPath = path
            return True
        except Exception:
            return False


class DownloadProgressButton(PopupToolButton):
    downloadProgress = pyqtSignal(int)
    downloadFinished = pyqtSignal()
    cancelled = pyqtSignal()

    def __init__(self, path, stat, parent=None):
        super(DownloadProgressButton, self).__init__(
            icon=getIcon('download.png'),
            parent=parent, mode=QToolButton.InstantPopup)
        self.downloadProgress.connect(self.onProgress)
        self.downloadFinished.connect(self.onFinished)
        self.path = path
        self.stat = stat
        self.readBytes = 0
        self.task = None

        self.menu.addAction(iCancel(), self.onCancel)

    def onCancel(self):
        if self.task:
            try:
                self.task.cancel()
            except:
                pass

            self.cancelled.emit()

    def onProgress(self, read):
        statInfo = StatInfo(self.stat)

        self.readBytes = read
        size = sizeFormat(statInfo.totalSize) if statInfo.valid else \
            iUnknown()

        self.setToolTip('{path} (size: {size}): downloaded {dl}'.format(
            path=self.path,
            size=size,
            dl=sizeFormat(self.readBytes))
        )

    def onFinished(self):
        self.setToolTip('{path}: download finished'.format(path=self.path))


class IconSelector(QComboBox):
    """
    A simple icon selector. The selected icon is injected in the
    repository and a signal with its CID is emitted
    """
    iconSelected = pyqtSignal(str)
    emptyIconSelected = pyqtSignal()

    iconsList = [
        ':/share/icons/distributed.png',
        ':/share/icons/atom.png',
        ':/share/icons/atom-rpg.png',
        ':/share/icons/cerebro.png',
        ':/share/icons/ipfs-cube-64.png',
        ':/share/icons/ipfs-logo-128-white.png',
        ':/share/icons/code-fork.png',
        ':/share/icons/cube-blue.png',
        ':/share/icons/cube-orange.png',
        ':/share/icons/cubehouse.png',
        ':/share/icons/fragments.png',
        ':/share/icons/go-home.png',
        ':/share/icons/hotspot.png',
        ':/share/icons/ipld-logo.png',
        ':/share/icons/pyramid-aqua.png',
        ':/share/icons/pyramid-stack.png',
        ':/share/icons/multimedia.png',
        ':/share/icons/folder-documents.png',
        ':/share/icons/folder-pictures.png',
        ':/share/icons/orbitdb.png',
        ':/share/icons/pyramid-hierarchy.png',
        ':/share/icons/sweethome.png',
        ':/share/icons/stroke-code.png',
        ':/share/icons/web-devel.png',
        ':/share/icons/mimetypes/image-x-generic.png',
        ':/share/icons/mimetypes/application-pdf.png',
        ':/share/icons/mimetypes/application-epub+zip.png',
        ':/share/icons/mimetypes/application-x-directory.png',
        ':/share/icons/mimetypes/text-html.png'
    ]

    def __init__(self, parent=None, offline=False, allowEmpty=False):
        super(IconSelector, self).__init__(parent)

        self.app = QApplication.instance()
        self.allowEmpty = allowEmpty
        self.offline = offline
        self.iconCid = None

        if self.app.system == 'Linux':
            self.setIconSize(QSize(64, 64))

        if allowEmpty is True:
            self.addItem('No icon')

        for iconP in self.iconsList:
            icon = getIcon(iconP)
            if icon:
                self.addItem(icon, '', QVariant(iconP))

        self.view = QListView(self)
        self.setView(self.view)
        self.currentIndexChanged.connect(self.onIndexChanged)

        if allowEmpty is False:
            self.injectCurrent()

    def injectCurrent(self):
        self.injectIconFromIndex(self.currentIndex())

    def injectIconFromIndex(self, idx):
        iconPath = self.itemData(idx)
        if iconPath:
            ensure(self.injectQrcIcon(iconPath))

    def onIndexChanged(self, idx):
        if self.allowEmpty and idx == 0:
            self.emptyIconSelected.emit()
        else:
            self.injectIconFromIndex(idx)

    @ipfsOp
    async def injectQrcIcon(self, op, iconPath):
        _file = QFile(iconPath)
        if _file.exists():
            if not _file.open(QIODevice.ReadOnly):
                return False

            data = _file.readAll().data()

            entry = await op.addBytes(data, offline=self.offline)
            if entry:
                self.iconCid = entry['Hash']
                self.iconSelected.emit(self.iconCid)
                self.setToolTip(self.iconCid)
                return True


class IPFSWebView(QWebEngineView):
    def __init__(self, webProfile=None, parent=None):
        super(IPFSWebView, self).__init__(parent=parent)

        self.app = QApplication.instance()
        self.webProfile = webProfile if webProfile else \
            self.app.webProfiles['ipfs']
        self.setPage(QWebEnginePage(self.webProfile, self))

        self.setMinimumSize(QSize(
            self.app.desktopGeometry.width() / 8,
            self.app.desktopGeometry.height() / 8
        ))


class MarkdownView(IPFSWebView):
    pass


class MarkdownInputWidget(QWidget):
    """
    General purpose markdown input editor, with (almost) live
    preview in a qtwebengine view

    It supports the 'dweb' URL scheme, so when you type something
    like ![img](dweb:/pathtoimage) the image should appear in
    the preview widget
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.app = QApplication.instance()

        self.markdownDocButton = QPushButton(
            'Have you completely lost your Markdown ?')
        self.markdownDocButton.setMaximumWidth(600)

        self.markdownDocButton.clicked.connect(self.onMarkdownHelp)
        helpLayout = QHBoxLayout()
        labelsLayout = QHBoxLayout()
        labelsLayout.addWidget(QLabel('Markdown input'))
        labelsLayout.addWidget(QLabel('Preview'))

        helpLayout.addWidget(self.markdownDocButton)
        helpLayout.addItem(
            QSpacerItem(10, 10, QSizePolicy.Expanding, QSizePolicy.Minimum))

        self.textEditUser = QTextEdit(self)
        self.textEditUser.textChanged.connect(self.onEdited)
        self.textEditMarkdown = MarkdownView(parent=self)

        self.updateTimeoutMs = 500
        self.updateTimer = QTimer()
        self.updateTimer.timeout.connect(self.onTimerOut)

        mainLayout = QVBoxLayout()
        editLayout = QHBoxLayout()
        editLayout.addWidget(self.textEditUser)
        editLayout.addWidget(self.textEditMarkdown)

        mainLayout.addLayout(labelsLayout)
        mainLayout.addLayout(editLayout)
        mainLayout.addLayout(helpLayout)

        self.setLayout(mainLayout)
        self.setMinSize()

    def onMarkdownHelp(self):
        ref = self.app.ipfsCtx.resources.get('markdown-reference')
        if ref:
            self.app.mainWindow.addBrowserTab().browseIpfsHash(
                ref['Hash']
            )

    def setMinSize(self):
        self.textEditUser.setMinimumSize(QSize(
            self.app.desktopGeometry.width() / 3,
            self.app.desktopGeometry.height() / 2)
        )
        self.textEditMarkdown.setMinimumSize(QSize(
            self.app.desktopGeometry.width() / 3,
            self.app.desktopGeometry.height() / 2)
        )

    def markdownText(self):
        return self.textEditUser.toPlainText()

    def onEdited(self):
        if self.updateTimer.isActive():
            self.updateTimer.stop()

        self.updateTimer.start(self.updateTimeoutMs)

    def onTimerOut(self):
        textData = self.textEditUser.toPlainText()
        try:
            html = markitdown(textData)
            self.textEditMarkdown.setHtml(html)
        except Exception:
            pass


class AtomFeedsToolbarButton(QToolButton):
    subscribeResult = pyqtSignal(str, int)

    def __init__(self, parent=None):
        super(AtomFeedsToolbarButton, self).__init__(parent=parent)
        self.app = QApplication.instance()
        self.model = self.app.modelAtomFeeds
        self.model.root.tracker.unreadCountChanged.connect(
            self.unreadEntriesCountChanged)
        self.setIcon(getIcon('atom-feed.png'))
        self.subscribeResult.connect(self.onSubResult)

        self.setToolTip('Atom Feeds reader')

    def unreadEntriesCountChanged(self, count):
        self.setToolTip('Atom Feeds reader: {count} unread entries'.format(
            count=count))

    def onSubResult(self, url, result):
        if result == 1:
            messageBox('This Atom feed is already registered', title=url)
        elif result == 2:
            messageBox('Error while subscribing to Atom feed', title=url)
        elif result == 3:
            messageBox('Subscribe OK!', title=url)

    async def atomFeedSubscribe(self, url):
        from galacteek.dweb.atom import AtomFeedExistsError
        try:
            await self.app.sqliteDb.feeds.follow(url)
        except AtomFeedExistsError:
            self.subscribeResult.emit(url, 1)
        except Exception:
            self.subscribeResult.emit(url, 2)
        else:
            self.subscribeResult.emit(url, 3)


class PageSourceWidget(QTextBrowser):
    def __init__(self, parent):
        super(PageSourceWidget, self).__init__(parent)

        self.app = QApplication.instance()
        self.setAcceptRichText(True)
        self.setLineWrapMode(QTextEdit.WidgetWidth)

    async def showSource(self, sourceText):
        formatter = HtmlFormatter()
        cssDefs = formatter.get_style_defs()

        lexer = get_lexer_by_name('html')

        self.document().setPlainText('Loading ...')
        try:
            high = await self.app.loop.run_in_executor(
                self.app.executor, highlight,
                sourceText, lexer, formatter)

            self.document().setDefaultStyleSheet(cssDefs)
            self.document().setHtml(high)
        except:
            self.document().setPlainText('Could not load source')

        self.moveCursor(QTextCursor.Start)


class AnimatedButton(QPushButton):
    """
    Animated button.
    """

    def __init__(self, clip, ignoreFrameEvery=1, parent=None):
        super().__init__(parent)

        self.clip = clip
        self.clip.finished.connect(self.clip.start)
        self.clip.frameChanged.connect(self.onFrameChanged)
        self.ignoreFrameEvery = ignoreFrameEvery

    def startClip(self):
        if not self.clip.playing():
            self.clip.start()

    def onFrameChanged(self, fNum):
        if divmod(fNum, self.ignoreFrameEvery)[1] == 0:
            self.setIcon(self.clip.createIcon())


class AnimatedLabel(QLabel):
    """
    Animated label.
    """

    animationClicked = pyqtSignal()

    def __init__(self, clip, loop=True, parent=None):
        super().__init__(parent)

        self.clip = clip

        if self.clip.isValid():
            self.setMovie(self.clip)

        if loop:
            self.clip.finished.connect(self.clip.start)

    def mousePressEvent(self, event):
        self.animationClicked.emit()

    def startClip(self):
        if not self.clip.playing():
            self.clip.start()



import os
import os.path
import asyncio
import functools
import aioipfs

from PyQt5.QtWidgets import QApplication
from PyQt5.QtCore import QCoreApplication
from PyQt5.QtCore import QObject
from PyQt5.QtCore import pyqtSignal

from galacteek import log
from galacteek import logUser
from galacteek import ensure

from galacteek.dweb.page import PDFViewerPage
from galacteek.dweb.page import WebTab
from galacteek.dweb.page import DWebView

from galacteek.ipfs import megabytes
from galacteek.ipfs import ipfsOp
from galacteek.ipfs.stat import StatInfo
from galacteek.ipfs.mimetype import MIMEType
from galacteek.ipfs.mimetype import detectMimeType
from galacteek.ipfs.mimetype import detectMimeTypeFromBuffer
from galacteek.ipfs.mimetype import mimeTypeDagUnknown

from galacteek.ipfs.cidhelpers import IPFSPath

from .dag import DAGViewer
from .textedit import TextEditorTab
from .dialogs import ResourceOpenConfirmDialog
from .helpers import getIcon
from .helpers import getMimeIcon
from .helpers import messageBox
from .helpers import runDialog
from .imgview import ImageViewerTab
from .i18n import iDagViewer


def iResourceCannotOpen(path):
    return QCoreApplication.translate(
        'resourceOpener',
        '{}: cannot determine resource type or timeout occured '
        '(check connectivity)').format(
            path)


class IPFSResourceOpener(QObject):
    objectOpened = pyqtSignal(IPFSPath)

    # Emitted by the opener when we want to show a dialog for confirmation
    needUserConfirm = pyqtSignal(IPFSPath, MIMEType, bool)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.app = QApplication.instance()
        self.setObjectName('resourceOpener')

        self.needUserConfirm.connect(self.onNeedUserConfirm)

    @ipfsOp
    async def open(self, ipfsop, pathRef,
                   mimeType=None,
                   openingFrom=None,
                   tryDecrypt=False,
                   fromEncrypted=False,
                   burnAfterReading=False):
        """
        Open the resource referenced by rscPath according
        to its MIME type

        :param pathRef: IPFS object's path (can be str or IPFSPath)
        :param MIMEType mimeType: MIME type
        """

        ipfsPath = None
        statInfo = None

        if isinstance(pathRef, IPFSPath):
            ipfsPath = pathRef
        elif isinstance(pathRef, str):
            ipfsPath = IPFSPath(pathRef, autoCidConv=True)
        else:
            raise ValueError('Invalid input value')

        if not ipfsPath.valid:
            return False

        rscPath = ipfsPath.objPath

        if self.app.mainWindow.pinAllGlobalChecked and not ipfsPath.isIpns:
            ensure(ipfsop.ctx.pinner.queue(rscPath, False,
                                           None,
                                           qname='default'))

        rscShortName = ipfsPath.shortRepr()

        if ipfsPath.isIpfs:
            # Try to reuse metadata from the multihash store
            rscMeta = await self.app.multihashDb.get(rscPath)
            if rscMeta:
                cachedMime = rscMeta.get('mimetype')
                cachedStat = rscMeta.get('stat')
                if cachedMime:
                    mimeType = MIMEType(cachedMime)
                if cachedStat:
                    statInfo = StatInfo(cachedStat)

        if mimeType is None:
            mimeType = await detectMimeType(rscPath)

        if mimeType and mimeType.valid:
            logUser.info('{path} ({type}): opening'.format(
                path=rscPath, type=str(mimeType)))
        else:
            logUser.info(iResourceCannotOpen(rscPath))
            return

        if mimeType == mimeTypeDagUnknown:
            indexPath = ipfsPath.child('index.html')
            stat = await ipfsop.objStat(indexPath.objPath, timeout=8)

            if stat:
                # Browse the index
                return self.app.mainWindow.addBrowserTab().browseFsPath(
                    indexPath)

            # Otherwise view the DAG
            view = DAGViewer(rscPath, self.app.mainWindow)
            self.app.mainWindow.registerTab(
                view, iDagViewer(),
                current=True,
                icon=getIcon('ipld.png'),
                tooltip=rscPath
            )
            return

        if mimeType.type == 'application/octet-stream' and not fromEncrypted:
            # Try to decode it with our key if it's a small file
            if statInfo is None:
                statInfo = StatInfo(await ipfsop.objStat(rscPath, timeout=5))

            profile = ipfsop.ctx.currentProfile
            if profile and statInfo.valid and \
                    (statInfo.dataSmallerThan(megabytes(8)) or tryDecrypt):
                data = await ipfsop.catObject(ipfsPath.objPath, timeout=30)
                if not data:
                    # XXX
                    return

                decrypted = await profile.rsaAgent.decrypt(data)

                if decrypted:
                    #
                    # "Good evening, 007"
                    #
                    # Create a short-lived IPFS offline file (not announced)
                    # with the decrypted content and open it
                    #

                    logUser.info('{path}: RSA OK'.format(path=rscPath))

                    # This one won't be announced or pinned
                    entry = await ipfsop.addBytes(decrypted,
                                                  offline=True,
                                                  pin=False)
                    if not entry:
                        logUser.info(
                            '{path}: cannot import decrypted file'.format(
                                path=rscPath))
                        return

                    # Open the decrypted file
                    return ensure(self.open(entry['Hash'],
                                            fromEncrypted=True,
                                            burnAfterReading=True))
                else:
                    logUser.debug(
                        '{path}: decryption impossible'.format(path=rscPath))

        elif mimeType.isText:
            tab = TextEditorTab(parent=self.app.mainWindow)
            tab.editor.display(ipfsPath)
            self.objectOpened.emit(ipfsPath)
            return self.app.mainWindow.registerTab(
                tab,
                rscShortName,
                icon=getMimeIcon('text/x-generic'),
                tooltip=rscPath,
                current=True
            )

        if mimeType.isImage or mimeType.isAnimation:
            tab = ImageViewerTab(self.app.mainWindow)
            ensure(tab.view.showImage(rscPath))
            self.objectOpened.emit(ipfsPath)
            return self.app.mainWindow.registerTab(
                tab,
                rscShortName,
                icon=getMimeIcon('image/x-generic'),
                tooltip=rscPath,
                current=True
            )

        if mimeType.isVideo or mimeType.isAudio:
            tab = self.app.mainWindow.addMediaPlayerTab()
            if tab:
                tab.playFromPath(rscPath)
            return

        if mimeType == 'application/pdf' and 0:  # not usable yet
            tab = WebTab(self.app.mainWindow)
            tab.attach(
                DWebView(page=PDFViewerPage(rscPath))
            )
            self.objectOpened.emit(ipfsPath)
            return self.app.mainWindow.registerTab(
                tab,
                rscShortName,
                icon=getMimeIcon('application/pdf'),
                tooltip=rscPath,
                current=True
            )

        if mimeType.isDir or ipfsPath.isIpns or mimeType.isHtml:
            self.objectOpened.emit(ipfsPath)
            return self.app.mainWindow.addBrowserTab().browseFsPath(ipfsPath)

        if openingFrom in ['filemanager', 'qa']:
            self.needUserConfirm.emit(ipfsPath, mimeType, True)
        else:
            self.needUserConfirm.emit(ipfsPath, mimeType, False)

    def onNeedUserConfirm(self, ipfsPath, mimeType, secureFlag):
        runDialog(ResourceOpenConfirmDialog, ipfsPath, mimeType, secureFlag,
                  accepted=functools.partial(self.onOpenConfirmed, ipfsPath,
                                             mimeType))

    def onOpenConfirmed(self, iPath, mType, dlg):
        return ensure(self.openWithSystemDefault(str(iPath)))

    @ipfsOp
    async def openBlock(self, ipfsop, pathRef, mimeType=None):
        """
        Open the raw block referenced by pathRef

        XXX: needs improvements, this only works for images for now
        """

        ipfsPath = None

        if isinstance(pathRef, IPFSPath):
            ipfsPath = pathRef
        elif isinstance(pathRef, str):
            ipfsPath = IPFSPath(pathRef)
        else:
            raise ValueError('Invalid input value')

        if not ipfsPath.valid:
            return False

        blockStat = await ipfsop.waitFor(
            ipfsop.client.block.stat(pathRef), 5
        )

        if not blockStat or not isinstance(blockStat, dict):
            messageBox('Block is bad')

        blockSize = blockStat.get('Size')

        if blockSize > megabytes(16):
            return

        logUser.info('Block {path}: Opening'.format(path=pathRef))

        blockData = await ipfsop.client.block.get(pathRef)

        rscPath = ipfsPath.objPath
        rscShortName = rscPath
        mimeType = await detectMimeTypeFromBuffer(blockData[:1024])

        if mimeType and mimeType.valid:
            logUser.info('Block: {path} ({type}): opening'.format(
                path=rscPath, type=str(mimeType)))
        else:
            return messageBox(iResourceCannotOpen(rscPath))

        if mimeType.isImage:
            tab = ImageViewerTab(self.app.mainWindow)
            ensure(tab.view.showImage(rscPath, fromBlock=True))
            self.objectOpened.emit(ipfsPath)
            return self.app.mainWindow.registerTab(
                tab,
                rscShortName,
                icon=getMimeIcon('image/x-generic'),
                tooltip=rscPath,
                current=True
            )

    @ipfsOp
    async def openWithExternal(self, ipfsop, rscPath, progArgs):
        filePath = os.path.join(self.app.tempDir.path(),
                                os.path.basename(rscPath))

        if not os.path.exists(filePath):
            try:
                await ipfsop.client.get(rscPath,
                                        dstdir=self.app.tempDir.path())
            except aioipfs.APIError as err:
                log.debug(err.message)
                return

        if not os.path.isfile(filePath):
            # Bummer
            return

        filePathEsc = filePath.replace('"', r'\"')
        args = progArgs.replace('%f', filePathEsc)

        log.debug('Object opener: executing: {}'.format(args))

        try:
            proc = await asyncio.create_subprocess_shell(
                args,
                stdout=asyncio.subprocess.PIPE
            )
            stdout, stderr = await proc.communicate()
        except BaseException as err:
            os.unlink(filePath)
            log.debug(str(err))
        else:
            os.unlink(filePath)

    @ipfsOp
    async def openWithSystemDefault(self, ipfsop, rscPath):
        # Use xdg-open or open depending on the platform
        if self.app.system == 'Linux':
            await self.openWithExternal(rscPath, 'xdg-open "%f"')
        elif self.app.system == 'Darwin':
            await self.openWithExternal(rscPath, 'open "%f"')



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'galacteek/ui/atomfeeds.ui'
#
# Created by: PyQt5 UI code generator 5.10
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_AtomFeeds(object):
    def setupUi(self, AtomFeeds):
        AtomFeeds.setObjectName("AtomFeeds")
        AtomFeeds.resize(503, 393)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(AtomFeeds)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.hLayout = QtWidgets.QHBoxLayout()
        self.hLayout.setContentsMargins(10, -1, 10, -1)
        self.hLayout.setSpacing(6)
        self.hLayout.setObjectName("hLayout")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.addFeedButton = QtWidgets.QPushButton(AtomFeeds)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/share/icons/atom-feed.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.addFeedButton.setIcon(icon)
        self.addFeedButton.setObjectName("addFeedButton")
        self.horizontalLayout.addWidget(self.addFeedButton)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.verticalLayout_2.addLayout(self.horizontalLayout)
        self.treeFeeds = QtWidgets.QTreeView(AtomFeeds)
        self.treeFeeds.setObjectName("treeFeeds")
        self.verticalLayout_2.addWidget(self.treeFeeds)
        self.hLayout.addLayout(self.verticalLayout_2)
        self.horizontalLayout_2.addLayout(self.hLayout)

        self.retranslateUi(AtomFeeds)
        QtCore.QMetaObject.connectSlotsByName(AtomFeeds)

    def retranslateUi(self, AtomFeeds):
        _translate = QtCore.QCoreApplication.translate
        AtomFeeds.setWindowTitle(_translate("AtomFeeds", "Form"))
        self.addFeedButton.setText(_translate("AtomFeeds", "Subscribe to an Atom feed"))

from . import galacteek_rc



from PyQt5.QtGui import QMovie
from PyQt5.QtGui import QIcon


class BaseClip(QMovie):
    def __init__(self, clipname='rotating-cube.gif', cache=True, speed=100):
        super(BaseClip, self).__init__(
            ':share/clips/{}'.format(clipname))
        self.setSpeed(speed)

        self.setCacheMode(
            QMovie.CacheAll if cache is True else QMovie.CacheNone)

    def createIcon(self):
        return QIcon(self.currentPixmap())

    def playing(self):
        return self.state() == QMovie.Running


class RotatingCubeClipSimple(BaseClip):
    def __init__(self, **kw):
        super().__init__(clipname='rotating-cube.gif', **kw)


class RotatingCubeRedFlash140d(BaseClip):
    def __init__(self, **kw):
        super().__init__(clipname='rotating-cube-redflash-140-6fps.gif', **kw)


class RotatingCubeClipFunky(BaseClip):
    def __init__(self, **kw):
        super().__init__(clipname='funky-cube-1.gif', **kw)



from async_generator import async_generator, yield_, yield_from_
import asyncio
import re

from PyQt5.QtGui import QStandardItem
from PyQt5.QtCore import QModelIndex
from PyQt5.QtCore import QObject
from PyQt5.QtCore import pyqtSignal

""" Helper functions to operate on QT item models """


def modelWalk(model, parent=QModelIndex(), search=None,
              searchre=None, columns=None, delete=False, dump=False,
              reflags=re.IGNORECASE, maxdepth=0, depth=0):
    """Walk over every model index in the given model looking for data matching
    the search parameter (or searchre if you want to use a regexp)

    Yields matching model indexes

    :param QAbstractItemModel model: the QT model
    :param QModelIndex parent: the parent item index in the model
    :param list columns: integer list of columns to search
    :param str search: string to search
    :param str searchre: regexp to search
    :param int reflags: regexp flags passed to re.search
    :param bool delete: delete matching indexes
    :param int maxdepth: maximum tree depth
    :return: list of matching indexes
    """

    for row in range(0, model.rowCount(parent)):
        index0 = model.index(row, 0, parent)
        if columns:
            cols = columns
        else:
            cols = [c for c in range(0, model.columnCount(parent))]
        for col in cols:
            index = model.index(row, col, parent)
            if not index:
                continue
            data = model.data(index)
            if search:
                if search == str(data):
                    if delete:
                        model.removeRow(row, parent)
                        model.submit()
                    yield index
            if searchre:
                if re.search(searchre, str(data), reflags):
                    yield index
        if model.hasChildren(index0):
            if maxdepth == 0 or (maxdepth > depth):
                yield from modelWalk(
                    model, parent=index0,
                    search=search, columns=columns, delete=delete,
                    dump=dump, searchre=searchre, reflags=reflags,
                    maxdepth=maxdepth, depth=depth)
                depth += 1


def modelSearch(model, parent=QModelIndex(), columns=None, delete=False,
                search=None, searchre=None, reflags=re.IGNORECASE):
    """Searches data in a QT model, see modelWalk for more info on
    parameters"""
    return list(modelWalk(model, parent=parent, columns=columns,
                          search=search, delete=delete, searchre=searchre,
                          reflags=reflags))


def modelDelete(model, search):
    """Delete items matching search from a QT model """
    return list(modelSearch(model, search=search, delete=True))


""" Asynchronous equivalents for when calling from coroutines """


@async_generator
async def modelWalkAsync(model, parent=QModelIndex(), search=None,
                         columns=None, delete=False, dump=False,
                         maxdepth=0, depth=0):
    for row in range(0, model.rowCount(parent)):
        await asyncio.sleep(0)

        index0 = model.index(row, 0, parent)
        if columns:
            cols = columns
        else:
            cols = [c for c in range(0, model.columnCount(parent))]
        for col in cols:
            index = model.index(row, col, parent)
            if not index:
                continue
            data = model.data(index)
            if search:
                if search == data:
                    if delete:
                        model.removeRow(row, parent)
                        model.submit()
                    await yield_(index)
        await asyncio.sleep(0)
        if model.hasChildren(index0):
            if maxdepth == 0 or (maxdepth > depth):
                await yield_from_(modelWalkAsync(
                    model, parent=index0,
                    search=search, columns=columns, delete=delete,
                    dump=dump, maxdepth=maxdepth, depth=depth))
                depth += 1


async def modelSearchAsync(model, parent=QModelIndex(), search=None,
                           columns=None, delete=False, maxdepth=0, depth=0):
    items = []
    async for v in modelWalkAsync(model, parent=parent, columns=columns,
                                  search=search, delete=delete,
                                  maxdepth=maxdepth, depth=depth):
        items.append(v)
    return items


async def modelDeleteAsync(model, search):
    return await modelSearchAsync(model, search=search, delete=True)


class UneditableItem(QStandardItem, QObject):
    refreshes = pyqtSignal()

    def __init__(self, text, icon=None):
        if icon:
            super(UneditableItem, self).__init__(icon, text)
        else:
            super(UneditableItem, self).__init__(text)
        self.setEditable(False)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'galacteek/ui/qschemecreatemapping.ui'
#
# Created by: PyQt5 UI code generator 5.10
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_QSchemeMappingDialog(object):
    def setupUi(self, QSchemeMappingDialog):
        QSchemeMappingDialog.setObjectName("QSchemeMappingDialog")
        QSchemeMappingDialog.resize(592, 297)
        self.gridLayout_3 = QtWidgets.QGridLayout(QSchemeMappingDialog)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.gridLayout_2 = QtWidgets.QGridLayout()
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.label = QtWidgets.QLabel(QSchemeMappingDialog)
        self.label.setObjectName("label")
        self.gridLayout_2.addWidget(self.label, 0, 0, 1, 1)
        self.mappingName = QtWidgets.QLineEdit(QSchemeMappingDialog)
        self.mappingName.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.mappingName.setMaxLength(32)
        self.mappingName.setObjectName("mappingName")
        self.gridLayout_2.addWidget(self.mappingName, 0, 1, 1, 1)
        self.label_2 = QtWidgets.QLabel(QSchemeMappingDialog)
        self.label_2.setObjectName("label_2")
        self.gridLayout_2.addWidget(self.label_2, 1, 0, 1, 1)
        self.ipnsResolveFrequency = QtWidgets.QSpinBox(QSchemeMappingDialog)
        self.ipnsResolveFrequency.setMaximumSize(QtCore.QSize(16777215, 32))
        self.ipnsResolveFrequency.setMinimum(60)
        self.ipnsResolveFrequency.setMaximum(86400)
        self.ipnsResolveFrequency.setProperty("value", 3600)
        self.ipnsResolveFrequency.setObjectName("ipnsResolveFrequency")
        self.gridLayout_2.addWidget(self.ipnsResolveFrequency, 1, 1, 1, 1)
        self.buttonBox = QtWidgets.QDialogButtonBox(QSchemeMappingDialog)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.gridLayout_2.addWidget(self.buttonBox, 2, 1, 1, 1)
        self.gridLayout_3.addLayout(self.gridLayout_2, 1, 0, 1, 1)
        self.mappedPath = QtWidgets.QLabel(QSchemeMappingDialog)
        self.mappedPath.setMaximumSize(QtCore.QSize(16777215, 32))
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.mappedPath.setFont(font)
        self.mappedPath.setText("")
        self.mappedPath.setObjectName("mappedPath")
        self.gridLayout_3.addWidget(self.mappedPath, 0, 0, 1, 1)

        self.retranslateUi(QSchemeMappingDialog)

    def retranslateUi(self, QSchemeMappingDialog):
        _translate = QtCore.QCoreApplication.translate
        QSchemeMappingDialog.setWindowTitle(_translate("QSchemeMappingDialog", "q:// scheme mapping"))
        self.label.setText(_translate("QSchemeMappingDialog", "Mapping name"))
        self.label_2.setText(_translate("QSchemeMappingDialog", "IPNS resolve frequency (seconds)"))




from PyQt5.QtWidgets import QApplication

from PyQt5.QtGui import QStandardItemModel

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtCore import QModelIndex
from PyQt5.QtCore import QObject

from galacteek import ensure
from galacteek.core.modelhelpers import *
from galacteek.ui.i18n import *  # noqa
from galacteek.ui.helpers import *

from galacteek.ui.i18n import iUnknown


class BaseItem(UneditableItem):
    def __init__(self, text, icon=None):
        super(BaseItem, self).__init__(text, icon=icon)

    def childrenItems(self):
        for row in range(0, self.rowCount()):
            child = self.child(row, 0)
            if child:
                yield child

    def findChildByName(self, name):
        for item in self.childrenItems():
            if not isinstance(item, str):
                continue
            if item.entry['Name'] == name:
                return item


class AtomFeedEntryItem(BaseItem):
    def __init__(self, entry, icon=None):
        self.entry = entry
        super(AtomFeedEntryItem, self).__init__(self.entry.title,
                                                icon=icon)

        self.setToolTip(self.entry.id)


class AtomFeedItem(BaseItem):
    unreadEntriesCounted = pyqtSignal(int)

    def __init__(self, feedid, title, icon=None,
                 offline=False, alwaysOffline=False):
        super(AtomFeedItem, self).__init__(title,
                                           icon=icon)

        self.feedId = feedid
        self.feedTitle = title
        self.unreadCount = 0
        self.setToolTip(self.feedId)

    def findEntryById(self, entryId):
        for item in self.childrenItems():
            if not isinstance(item, AtomFeedEntryItem):
                continue
            if item.entry.id == entryId:
                return item

    def updateTitle(self):
        unreadCount = 0
        for item in self.childrenItems():
            if not isinstance(item, AtomFeedEntryItem):
                # should not happen
                continue

            if item.entry.status == item.entry.ENTRY_STATUS_NEW:
                unreadCount += 1

        self.setText('{title} ({count})'.format(
            title=self.feedTitle,
            count=unreadCount
        ))

        if unreadCount != self.unreadCount:
            self.unreadCount = unreadCount
            self.parent().update()


class RootItemTracker(QObject):
    unreadCountChanged = pyqtSignal(int)


class RootItem(BaseItem):
    def __init__(self, text, icon=None):
        super(RootItem, self).__init__(text, icon=icon)

        self.unreadEntriesCount = 0
        self.tracker = RootItemTracker()

    def findFeedById(self, feedId):
        for item in self.childrenItems():
            if not isinstance(item, AtomFeedItem):
                continue
            if item.feedId == feedId:
                return item

    def update(self):
        self.unreadEntriesCount = self.unreadEntriesCalculate()
        self.tracker.unreadCountChanged.emit(self.unreadEntriesCount)

    def unreadEntriesCalculate(self):
        total = 0
        for item in self.childrenItems():
            if isinstance(item, AtomFeedItem):
                total += item.unreadCount
        return total


class AtomFeedsModel(QStandardItemModel):
    # An item's view was refreshed
    needExpand = pyqtSignal(QModelIndex)
    feedEntryAdded = pyqtSignal(AtomFeedEntryItem)

    def __init__(self, feedsDb, parent=None):
        super(AtomFeedsModel, self).__init__(parent)
        self.app = QApplication.instance()
        self.db = feedsDb

        self.db.processedFeedEntry.connect(self.onProcessedEntry)
        self.db.feedRemoved.connect(self.onFeedRemoved)

        self.itemRoot = RootItem('Root')

        self.invisibleRootItem().appendRow(self.itemRoot)
        self.itemRootIdx = self.indexFromItem(self.itemRoot)

        self.setHorizontalHeaderLabels(['Title', 'Date'])

    @property
    def root(self):
        return self.itemRoot

    def markEntryAsRead(self, entryItem):
        entryItem.entry.status = entryItem.entry.ENTRY_STATUS_READ
        if entryItem.entry.srow_id:
            ensure(self.db.feedEntrySetStatus(entryItem.entry.srow_id,
                                              entryItem.entry.status))

    def onFeedRemoved(self, feedId):
        child = self.itemRoot.findFeedById(feedId)
        if child:
            self.itemRoot.removeRow(child.row())

    def onProcessedEntry(self, feed, entry):
        child = self.itemRoot.findFeedById(feed.id)
        if not child:
            feedItem = AtomFeedItem(feed.id, feed.title)
            self.itemRoot.appendRow(feedItem)
        else:
            feedItem = child

        entryFound = feedItem.findEntryById(entry.id)
        if not entryFound:
            entryItem = AtomFeedEntryItem(entry)
            date = entry.published

            dateItem = BaseItem(
                date.strftime('%Y-%m-%d %H:%M') if date else iUnknown()
            )
            feedItem.appendRow([
                entryItem,
                dateItem
            ])

            self.feedEntryAdded.emit(entryItem)



import os.path
import asyncio
import functools
import async_timeout

from PyQt5.QtWidgets import QApplication
from PyQt5.QtWidgets import QAction
from PyQt5.QtWidgets import QActionGroup
from PyQt5.QtWidgets import QFileDialog
from PyQt5.QtWidgets import QTreeView
from PyQt5.QtWidgets import QMenu
from PyQt5.QtWidgets import QAbstractItemView
from PyQt5.QtWidgets import QInputDialog
from PyQt5.QtWidgets import QWidget
from PyQt5.QtWidgets import QFileSystemModel
from PyQt5.QtWidgets import QHeaderView

from PyQt5.QtCore import QCoreApplication
from PyQt5.QtCore import Qt
from PyQt5.QtCore import QDir
from PyQt5.QtCore import QSize
from PyQt5.QtCore import pyqtSignal

from galacteek import ensure
from galacteek import GALACTEEK_NAME
from galacteek.ipfs.cidhelpers import joinIpfs
from galacteek.ipfs.cidhelpers import IPFSPath
from galacteek.ipfs.cidhelpers import cidConvertBase32
from galacteek.ipfs.ipfsops import *
from galacteek.ipfs.wrappers import ipfsOp
from galacteek.ipfs.mimetype import MIMEType
from galacteek.appsettings import *

from galacteek.core import modelhelpers
from galacteek.core.models.mfs import MFSItem
from galacteek.core.models.mfs import MFSNameItem
from galacteek.core.models.mfs import MFSRootItem

from . import ui_files
from . import dag
from .i18n import *  # noqa
from .helpers import *  # noqa
from .widgets import GalacteekTab
from .hashmarks import *  # noqa
from .clipboard import iCopyCIDToClipboard
from .clipboard import iCopyPathToClipboard

import aioipfs

# Files messages


def iFileImportError():
    return QCoreApplication.translate(
        'FileManagerForm', 'Error importing file {}')


def iAddedFile(name):
    return QCoreApplication.translate('FileManagerForm',
                                      'Added {0}').format(name)


def iImportedCount(count):
    return QCoreApplication.translate('FileManagerForm',
                                      'Imported {0} file(s)').format(count)


def iLoadingFile(name):
    return QCoreApplication.translate('FileManagerForm',
                                      'Loading file {0}').format(name)


def iLoading(name):
    return QCoreApplication.translate('FileManagerForm',
                                      'Loading {0}').format(name)


def iListingMFSPath(path):
    return QCoreApplication.translate(
        'FileManagerForm',
        'Listing MFS path: {0}').format(path)


def iListingMFSTimeout(path):
    return QCoreApplication.translate(
        'FileManagerForm',
        'Timeout while listing MFS path: {0}').format(path)


def iOpenWith():
    return QCoreApplication.translate('FileManagerForm', 'Open with')


def iDeleteFile():
    return QCoreApplication.translate('FileManagerForm', 'Delete file')


def iExploreDir():
    return QCoreApplication.translate('FileManagerForm', 'Explore directory')


def iUnlinkFile():
    return QCoreApplication.translate('FileManagerForm', 'Unlink file')


def iHashmarkFile():
    return QCoreApplication.translate('FileManagerForm', 'Hashmark')


def iBrowseFile():
    return QCoreApplication.translate('FileManagerForm', 'Browse')


def iSelectDirectory():
    return QCoreApplication.translate('FileManagerForm', 'Select directory')


def iSelectFiles():
    return QCoreApplication.translate('FileManagerForm',
                                      'Select one or more files to import')


def iOfflineMode():
    return QCoreApplication.translate('FileManagerForm', 'Offline mode')


def iChunker():
    return QCoreApplication.translate('FileManagerForm', 'Chunker')


def iDhtProvide():
    return QCoreApplication.translate(
        'FileManagerForm',
        'Announce (DHT provide)')


def iDhtProvideRecursive():
    return QCoreApplication.translate(
        'FileManagerForm',
        'Announce (DHT provide, recursive)')


def iRawBlocksForLeaves():
    return QCoreApplication.translate(
        'FileManagerForm',
        'Use raw blocks for leaf nodes')


def iDAGGenerationFormat():
    return QCoreApplication.translate(
        'FileManagerForm',
        'DAG generation format')


def iDAGFormatBalanced():
    return QCoreApplication.translate(
        'FileManagerForm',
        'Balanced')


def iDAGFormatTrickle():
    return QCoreApplication.translate(
        'FileManagerForm',
        'Trickle')


def iOfflineModeToolTip():
    return QCoreApplication.translate(
        'FileManagerForm',
        'Offline mode: in this mode your files will not be '
        'immediately announced on the DHT')


class FileManager(QWidget):
    statusReady = 0
    statusBusy = 1

    displayedItemChanged = pyqtSignal(MFSNameItem)

    def __init__(self, parent=None):
        super().__init__(parent)

        self.app = QApplication.instance()
        self.clipboard = self.app.appClipboard

        self.lock = asyncio.Lock()
        self.model = None
        self._offlineMode = False

        self.ui = ui_files.Ui_FileManagerForm()
        self.ui.setupUi(self)

        self.status = self.statusReady

        # Build file browser
        self.displayedItem = None
        self.createFileManager()

        self.filesDialog = QFileDialog(None)
        self.dialogLastDirSelected = None

        self.ui.localFileManagerSwitch.setCheckable(True)

        # Connect the various buttons
        self.ui.helpButton.clicked.connect(
            lambda: self.app.manuals.browseManualPage('filemanager.html'))
        self.ui.addFileButton.clicked.connect(self.onAddFilesClicked)
        self.ui.addDirectoryButton.clicked.connect(self.onAddDirClicked)
        self.ui.refreshButton.clicked.connect(self.onRefreshClicked)
        self.ui.searchFiles.returnPressed.connect(self.onSearchFiles)
        self.ui.localFileManagerSwitch.toggled.connect(
            self.onLocalFileManagerToggled)
        self.ui.fileManagerButton.clicked.connect(self.onLocalFileManager)

        # FS options button
        fsOptsMenu = QMenu(self)
        fsDagFormatMenu = QMenu(iDAGGenerationFormat(), self)
        fsDagFormatMenu.setIcon(getIcon('ipld.png'))

        fsChunkerMenu = self.buildChunkerMenu()

        fsMiscOptsMenu = QMenu('Options', self)
        fsMiscOptsMenu.setIcon(getIcon('folder-black.png'))

        self.rawLeavesAction = QAction(
            iRawBlocksForLeaves(), fsMiscOptsMenu)
        self.rawLeavesAction.setCheckable(True)

        fsMiscOptsMenu.addAction(self.rawLeavesAction)

        self.dagFormatGroup = QActionGroup(fsDagFormatMenu)
        self.dagFormatGroup.triggered.connect(self.onDagFormatChanged)

        self.dagFormatBalancedAction = QAction(
            iDAGFormatBalanced(), self.dagFormatGroup)
        self.dagFormatBalancedAction.setCheckable(True)
        self.dagFormatBalancedAction.setChecked(True)
        self.dagFormatTrickleAction = QAction(
            iDAGFormatTrickle(), self.dagFormatGroup)
        self.dagFormatTrickleAction.setCheckable(True)

        self.dagFormatGroup.addAction(self.dagFormatBalancedAction)
        self.dagFormatGroup.addAction(self.dagFormatTrickleAction)

        fsDagFormatMenu.addActions(self.dagFormatGroup.actions())

        fsOptsMenu.addMenu(fsDagFormatMenu)
        fsOptsMenu.addSeparator()
        fsOptsMenu.addMenu(fsChunkerMenu)
        fsOptsMenu.addSeparator()
        fsOptsMenu.addMenu(fsMiscOptsMenu)

        self.ui.fsOptionsButton.setPopupMode(QToolButton.InstantPopup)
        self.ui.fsOptionsButton.setMenu(fsOptsMenu)

        # Offline mode button
        self.ui.offlineButton.setObjectName('filemanagerOfflineButton')
        self.ui.offlineButton.setText(iOfflineMode())
        self.ui.offlineButton.setCheckable(True)
        self.ui.offlineButton.toggled.connect(self.onOfflineToggle)
        self.ui.offlineButton.setChecked(False)

        # Connect the tree view actions
        self.ui.treeFiles.doubleClicked.connect(self.onDoubleClicked)
        self.ui.treeFiles.setContextMenuPolicy(Qt.CustomContextMenu)
        self.ui.treeFiles.customContextMenuRequested.connect(
            self.onContextMenu)
        self.ui.treeFiles.expanded.connect(self.onExpanded)
        self.ui.treeFiles.collapsed.connect(self.onCollapsed)

        self.ui.comboIconSize.addItem('Small')
        self.ui.comboIconSize.addItem('Medium')
        self.ui.comboIconSize.addItem('Large')
        self.ui.comboIconSize.currentIndexChanged.connect(self.onIconSize)

        self.ui.collapseAll.clicked.connect(self.onCollapseAll)
        self.ui.collapseAll.hide()

        # Connect the event filter
        evfilter = IPFSTreeKeyFilter(self.ui.treeFiles)
        evfilter.copyHashPressed.connect(self.onCopyItemHash)
        evfilter.copyPathPressed.connect(self.onCopyItemPath)
        evfilter.returnPressed.connect(self.onReturn)
        evfilter.explorePressed.connect(self.onExploreItem)
        self.ui.treeFiles.installEventFilter(evfilter)

        self.displayedItemChanged.connect(self.onItemChange)

        # Setup the tree view
        self.ui.treeFiles.setExpandsOnDoubleClick(True)
        self.ui.treeFiles.setItemsExpandable(True)
        self.ui.treeFiles.setSortingEnabled(True)
        self.ui.treeFiles.sortByColumn(0, Qt.AscendingOrder)
        self.ui.treeFiles.setIconSize(QSize(16, 16))

        self.iconFolder = getIcon('folder-open.png')
        self.iconFile = getIcon('file.png')

        # Configure drag-and-drop
        self.ui.treeFiles.setAcceptDrops(True)
        self.ui.treeFiles.setDragDropMode(QAbstractItemView.DragDrop)

        self.ipfsKeys = []

    @property
    def gWindow(self):
        return self.app.mainWindow

    @property
    def profile(self):
        return self.app.ipfsCtx.currentProfile

    @property
    def offlineMode(self):
        if self.displayedItem:
            return self.displayedItem.offline

        return self._offlineMode

    @property
    def dagGenerationFormat(self):
        cAction = self.dagFormatGroup.checkedAction()
        if cAction is self.dagFormatBalancedAction:
            return 'balanced'
        elif cAction is self.dagFormatTrickleAction:
            return 'trickle'
        return 'unknown'

    @property
    def useRawLeaves(self):
        return self.rawLeavesAction.isChecked()

    @property
    def displayPath(self):
        if self.displayedItem:
            return self.displayedItem.path

    @property
    def busy(self):
        return self.status == self.statusBusy

    @property
    def rscOpenTryDecrypt(self):
        return self.displayedItem is self.model.itemEncrypted or \
            self.displayedItem is self.model.itemQrCodes

    @property
    def inEncryptedFolder(self):
        return self.displayedItem is self.model.itemEncrypted

    def buildChunkerMenu(self):
        # Build the menu to select the chunking strategy
        fsChunkerMenu = QMenu(iChunker(), self)

        self.chunkingAlg = 'size-262144'
        self.chunkerGroup = QActionGroup(fsChunkerMenu)
        self.chunkerGroup.triggered.connect(self.onChunkerChanged)

        # Fixed block size algorithm
        sizes = [int(32768 * x) for x in range(4, 32, 4)]

        for size in sizes:
            action = QAction('Fixed-size chunker (block size: {} kb)'.format(
                int(size / 1024)), self.chunkerGroup)
            action.setData('size-{}'.format(size))
            action.setCheckable(True)

            # Check the default chunker
            if size == 262144:
                action.setChecked(True)

        # Rabin.
        # This is just a small selection of block sizes people might
        # want to try. For each minimum block size we calculate a number
        # of maximum block sizes (the average is always the median value)

        msizes = [int(32768 * x) for x in range(2, 16, 2)]

        for msize in msizes:
            maxsizes = [int(msize * z) for z in range(2, 6, 1)]

            for maxs in maxsizes:
                if maxs >= (1024 * 1024):
                    break

                avgs = msize + int((maxs - msize) / 2)

                action = QAction(
                    'Rabin chunker '
                    '(min: {mbs} kb, avg: {avgbs} kb, max: {maxbs} kb)'.format(
                        mbs=int(msize / 1024),
                        avgbs=int(avgs / 1024),
                        maxbs=int(maxs / 1024)
                    ), self.chunkerGroup
                )

                action.setCheckable(True)
                action.setData('rabin-{0}-{1}-{2}'.format(msize, avgs, maxs))

        fsChunkerMenu.addActions(self.chunkerGroup.actions())
        return fsChunkerMenu

    def createFileManager(self):
        self.fManagerModel = QFileSystemModel()
        self.fManagerModel.setRootPath('')

        self.localTree = QTreeView()
        self.localTree.setModel(self.fManagerModel)
        self.localTree.setDragEnabled(True)
        self.localTree.setDragDropMode(QAbstractItemView.DragOnly)
        self.localTree.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.localTree.setItemsExpandable(True)

        rootIndex = self.fManagerModel.index(QDir.rootPath())
        if rootIndex.isValid():
            self.localTree.setRootIndex(rootIndex)

        if self.app.system in ['Linux', 'FreeBSD', 'Darwin']:
            """
            Expand the user's home folder and its parents
            """

            home = QDir(QDir.homePath())
            myHomeIndex = self.fManagerModel.index(home.path())

            if myHomeIndex.isValid():
                self.localTree.expand(myHomeIndex)

                while home.cdUp():
                    idx = self.fManagerModel.index(home.path())
                    if idx.isValid():
                        self.localTree.expand(idx)

        for col in range(1, 4):
            self.localTree.hideColumn(col)

        self.localTree.hide()
        self.ui.hLayoutBrowser.insertWidget(0, self.localTree)

    def setupPathSelector(self):
        def c():
            return self.ui.pathSelector.count()

        self.ui.pathSelector.clear()

        self.ui.pathSelector.setObjectName('fmanagerPathSelector')
        self.ui.pathSelector.insertItem(c(), self.model.itemHome.icon(),
                                        'Home')
        self.ui.pathSelector.insertSeparator(c())
        self.ui.pathSelector.insertItem(c(), self.model.itemImages.icon(),
                                        iImages())
        self.ui.pathSelector.insertItem(c(),
                                        self.model.itemPictures.icon(),
                                        iPictures())
        self.ui.pathSelector.insertItem(c(), self.model.itemVideos.icon(),
                                        iVideos())
        self.ui.pathSelector.insertItem(c(), self.model.itemMusic.icon(),
                                        iMusic())
        self.ui.pathSelector.insertSeparator(c())
        self.ui.pathSelector.insertItem(c(), self.model.itemCode.icon(),
                                        iCode())
        self.ui.pathSelector.insertItem(c(),
                                        self.model.itemDocuments.icon(),
                                        iDocuments())
        self.ui.pathSelector.insertSeparator(c())
        self.ui.pathSelector.insertItem(c(),
                                        self.model.itemWebPages.icon(),
                                        iWebPages())
        self.ui.pathSelector.insertItem(c(),
                                        self.model.itemDWebApps.icon(),
                                        iDWebApps())
        self.ui.pathSelector.insertSeparator(c())
        self.ui.pathSelector.insertItem(c(),
                                        self.model.itemQrCodes.icon(),
                                        iQrCodes())
        self.ui.pathSelector.insertSeparator(c())
        self.ui.pathSelector.insertItem(c(),
                                        self.model.itemTemporary.icon(),
                                        iTemporaryFiles())
        self.ui.pathSelector.insertSeparator(c())
        self.ui.pathSelector.insertItem(c(),
                                        self.model.itemEncrypted.icon(),
                                        iEncryptedFiles())
        self.ui.pathSelector.activated.connect(self.onPathSelector)

        if self.app.system == 'Linux':
            self.ui.pathSelector.setIconSize(QSize(24, 24))

    def setupModel(self):
        if self.profile is None:
            return

        # Setup the model if needed
        if self.model is None and self.profile.filesModel:
            self.model = self.profile.filesModel

            if not self.model.qrInitialized:
                ensure(self.model.setupQrCodesFolder(
                    self.profile, self.model, self))

        if not self.model:
            return

        self.ui.treeFiles.setModel(self.model)
        self.ui.treeFiles.header().setSectionResizeMode(
            QHeaderView.ResizeToContents)

        if self.displayedItem is None:
            self.changeDisplayItem(self.model.itemHome)

        self.app.task(self.updateKeys)

        self.disconnectDropSignals()

        # Connect the model's drag-and-drop signals
        self.model.fileDropEvent.connect(self.onDropFile)
        self.model.directoryDropEvent.connect(self.onDropDirectory)

        if self.ui.pathSelector.count() == 0:
            self.setupPathSelector()

    def enableButtons(self, flag=True):
        for btn in [self.ui.addFileButton,
                    self.ui.addDirectoryButton,
                    self.ui.refreshButton,
                    self.ui.localFileManagerSwitch,
                    self.ui.fileManagerButton,
                    self.ui.pathSelector,
                    self.ui.offlineButton,
                    self.ui.searchFiles]:
            btn.setEnabled(flag)

        if self.displayedItem is self.model.itemEncrypted:
            self.ui.addDirectoryButton.setEnabled(False)

    def currentItem(self):
        currentIdx = self.ui.treeFiles.currentIndex()
        if currentIdx.isValid():
            return self.model.getNameItemFromIdx(currentIdx)

    def onDagFormatChanged(self, action):
        pass

    def onChunkerChanged(self, action):
        self.chunkingAlg = action.data()

    def onIconSize(self, index):
        size = None

        if index == 0:
            size = QSize(16, 16)
        elif index == 1:
            size = QSize(24, 24)
        elif index == 2:
            size = QSize(32, 32)
        else:
            return

        self.ui.treeFiles.setIconSize(size)

    def onCollapseAll(self):
        self.ui.treeFiles.collapseAll()

        if self.displayedItem:
            self.displayedItem.expandedItemsCount = 0
            self.collapseButtonUpdate()

    def onOfflineToggle(self, checked):
        if self.displayedItem:
            self.displayedItem.offline = checked

        self.ui.offlineButton.setToolTip(iOfflineModeToolTip())

    def onClose(self):
        if not self.busy:
            self.disconnectDropSignals()
            return True
        return False

    def onItemChange(self, item):
        self.collapseButtonUpdate()

        if isinstance(item, MFSRootItem):
            self.ui.offlineButton.setChecked(item.offline)

    def collapseButtonUpdate(self):
        self.ui.collapseAll.setVisible(
            self.displayedItem.expandedItemsCount > 0)

    def disconnectDropSignals(self):
        try:
            self.model.fileDropEvent.disconnect(self.onDropFile)
            self.model.directoryDropEvent.disconnect(self.onDropDirectory)
        except Exception:
            pass

    def onLocalFileManagerToggled(self, checked):
        self.localTree.setVisible(checked)

    def onLocalFileManager(self):
        self.ui.localFileManagerSwitch.setChecked(
            not self.ui.localFileManagerSwitch.isChecked())

    def onDropFile(self, path):
        self.scheduleAddFiles([path])

    def onDropDirectory(self, path):
        self.scheduleAddDirectory(path)

    def onCopyItemHash(self):
        currentItem = self.currentItem()
        if currentItem:
            dataHash = self.model.getHashFromIdx(currentItem.index())
            self.app.setClipboardText(dataHash)

    def onCopyItemPath(self):
        currentItem = self.currentItem()
        if currentItem:
            nameItem = self.model.getNameItemFromIdx(currentItem.index())
            if nameItem:
                self.app.setClipboardText(nameItem.fullPath)

    def onReturn(self):
        currentItem = self.currentItem()
        dataHash = self.model.getHashFromIdx(currentItem.index())
        if dataHash:
            self.gWindow.addBrowserTab().browseIpfsHash(dataHash)

    def onExpanded(self, idx):
        if self.displayedItem and idx != self.displayedItem.index():
            self.displayedItem.expandedItemsCount += 1
            self.collapseButtonUpdate()

    def onCollapsed(self, idx):
        if self.displayedItem and idx != self.displayedItem.index():
            self.displayedItem.expandedItemsCount -= 1
            self.collapseButtonUpdate()

    def pathSelectorDefault(self):
        self.ui.pathSelector.setCurrentIndex(0)

    def onPathSelector(self, idx):
        if self.busy:
            return

        text = self.ui.pathSelector.itemText(idx)

        if text == iHome():
            self.changeDisplayItem(self.model.itemHome)
        elif text == iPictures():
            self.changeDisplayItem(self.model.itemPictures)
        elif text == iImages():
            self.changeDisplayItem(self.model.itemImages)
        elif text == iVideos():
            self.changeDisplayItem(self.model.itemVideos)
        elif text == iMusic():
            self.changeDisplayItem(self.model.itemMusic)
        elif text == iCode():
            self.changeDisplayItem(self.model.itemCode)
        elif text == iDocuments():
            self.changeDisplayItem(self.model.itemDocuments)
        elif text == iWebPages():
            self.changeDisplayItem(self.model.itemWebPages)
        elif text == iDWebApps():
            self.changeDisplayItem(self.model.itemDWebApps)
        elif text == iQrCodes():
            self.changeDisplayItem(self.model.itemQrCodes)
        elif text == iTemporaryFiles():
            self.changeDisplayItem(self.model.itemTemporary)
        elif text == iEncryptedFiles():
            self.changeDisplayItem(self.model.itemEncrypted)

        if self.app.settingsMgr.hideHashes:
            self.ui.treeFiles.hideColumn(2)

    def changeDisplayItem(self, item):
        self.displayedItem = item
        self.ui.treeFiles.setRootIndex(self.displayedItem.index())
        self.updateTree()
        self.ui.treeFiles.expand(self.displayedItem.index())

        if isinstance(item, MFSRootItem):
            self.displayedItemChanged.emit(item)

        if self.app.settingsMgr.hideHashes:
            self.ui.treeFiles.hideColumn(2)

    def onContextMenuVoid(self, point):
        menu = QMenu(self)
        menu.exec(self.ui.treeFiles.mapToGlobal(point))

    def onContextMenu(self, point):
        idx = self.ui.treeFiles.indexAt(point)
        if not idx.isValid():
            return self.onContextMenuVoid(point)

        nameItem = self.model.getNameItemFromIdx(idx)
        dataHash = self.model.getHashFromIdx(idx)
        ipfsPath = nameItem.fullPath
        menu = QMenu(self)

        def explore(cid):
            self.gWindow.explore(cid)

        def hashmark(mPath, name):
            addHashmark(self.app.marksLocal, mPath, name)

        def copyHashToClipboard(itemHash):
            self.clipboard.setText(itemHash)

        def openWithMediaPlayer(itemHash):
            parentHash = nameItem.parentHash
            name = nameItem.entry['Name']
            if parentHash:
                fp = joinIpfs(os.path.join(parentHash, name))
                self.gWindow.mediaPlayerPlay(fp, mediaName=name)
            else:
                self.gWindow.mediaPlayerPlay(joinIpfs(itemHash),
                                             mediaName=name)

        menu.addAction(getIcon('clipboard.png'),
                       iCopyCIDToClipboard(),
                       functools.partial(self.app.setClipboardText, dataHash))
        if nameItem.fullPath:
            menu.addAction(getIcon('clipboard.png'),
                           iCopyPathToClipboard(),
                           functools.partial(self.app.setClipboardText,
                                             ipfsPath))
        menu.addSeparator()
        menu.addAction(getIconIpfs64(), iDhtProvide(), functools.partial(
            self.onDhtProvide, dataHash, False))
        menu.addAction(getIconIpfs64(), iDhtProvideRecursive(),
                       functools.partial(self.onDhtProvide, dataHash, True))
        menu.addSeparator()

        menu.addAction(getIcon('ipld.png'), iDagView(),
                       functools.partial(self.onDagView, dataHash))
        menu.addSeparator()

        menu.addAction(getIcon('hashmarks.png'),
                       iHashmarkFile(),
                       functools.partial(hashmark, ipfsPath,
                                         nameItem.entry['Name']))
        menu.addAction(getIconIpfs64(),
                       iBrowseFile(),
                       functools.partial(self.browse, ipfsPath))
        menu.addAction(getIcon('open.png'),
                       iOpen(),
                       functools.partial(ensure, self.app.resourceOpener.open(
                           ipfsPath, openingFrom='filemanager',
                           tryDecrypt=self.rscOpenTryDecrypt)))
        if nameItem.isFile():
            menu.addAction(getIcon('text-editor.png'),
                           'Edit',
                           functools.partial(self.editFile,
                                             nameItem.ipfsPath))

        if nameItem.isDir():
            menu.addAction(getIcon('folder-open.png'),
                           iExploreDir(),
                           functools.partial(explore, dataHash))

        def publishToKey(action):
            key = action.data()['key']['Name']
            oHash = action.data()['hash']

            async def publish(op, oHash, keyName):
                await op.publish(joinIpfs(oHash), key=keyName)

            self.app.ipfsTaskOp(publish, oHash, key)

        # Delete/unlink
        menu.addSeparator()
        menu.addAction(iUnlinkFile(), functools.partial(
            self.scheduleUnlink, nameItem, dataHash))
        menu.addAction(iDeleteFile(), functools.partial(
            self.scheduleDelete, nameItem, dataHash))
        menu.addSeparator()

        # Populate publish menu
        publishMenu = QMenu('Publish to IPNS key', self)
        publishMenu.setIcon(getIcon('key-diago.png'))
        for key in self.ipfsKeys:
            if not key['Name'] or key['Name'].startswith(GALACTEEK_NAME):
                continue

            action = QAction(key['Name'], self)
            action.setData({
                'key': key,
                'hash': dataHash
            })

            publishMenu.addAction(action)

        publishMenu.triggered.connect(publishToKey)

        menu.addSeparator()
        menu.addMenu(publishMenu)
        menu.exec(self.ui.treeFiles.mapToGlobal(point))

    def browse(self, path):
        self.gWindow.addBrowserTab().browseFsPath(
            IPFSPath(path, autoCidConv=True))

    def onDagView(self, cid):
        view = dag.DAGViewer(joinIpfs(cid), self.app.mainWindow)
        self.app.mainWindow.registerTab(
            view, iDagViewer(),
            current=True,
            icon=getIcon('ipld.png'),
            tooltip=cid
        )

    def onDhtProvide(self, cid, recursive):
        @ipfsOpFn
        async def dhtProvide(ipfsop, value, recursive=True):
            if await ipfsop.provide(value, recursive=recursive) is True:
                logUser.info('{0}: DHT provide OK!'.format(value))
            else:
                logUser.info('{0}: DHT provide error'.format(value))

        ensure(dhtProvide(cid, recursive=recursive))

    def editFile(self, ipfsPath):
        self.gWindow.addEditorTab(path=ipfsPath)

    def onExploreItem(self):
        current = self.currentItem()

        if current and current.isDir():
            dataHash = self.model.getHashFromIdx(current.index())
            self.gWindow.explore(dataHash)

    def onDoubleClicked(self, idx):
        resourceOpener = self.gWindow.app.resourceOpener

        if not idx.isValid():
            return

        nameItem = self.model.getNameItemFromIdx(idx)
        item = self.model.itemFromIndex(idx)
        dataHash = self.model.getHashFromIdx(idx)

        if nameItem.isFile():
            fileName = nameItem.text()
            finalPath = joinIpfs(dataHash)

            # Find the parent hash
            parentHash = nameItem.parentHash
            if parentHash:
                # We have the parent hash, so use it to build a file path
                # preserving the real file name
                finalPath = joinIpfs(os.path.join(parentHash, fileName))

            ensure(resourceOpener.open(
                finalPath,
                openingFrom='filemanager',
                tryDecrypt=self.rscOpenTryDecrypt
            ))

        elif nameItem.isDir():
            self.app.task(self.listFiles, item.path, parentItem=item,
                          autoexpand=True)

    def onSearchFiles(self):
        search = self.ui.searchFiles.text()
        self.ui.treeFiles.keyboardSearch(search)

    def onMkDirClicked(self):
        dirName = QInputDialog.getText(self, 'Directory name',
                                       'Directory name')
        if dirName:
            self.app.task(self.makeDir, self.displayedItem.path,
                          dirName[0])

    def onRefreshClicked(self):
        self.updateTree()
        self.ui.treeFiles.setFocus(Qt.OtherFocusReason)

    def onAddDirClicked(self):
        dialog = QFileDialog(None)

        dialog.setFileMode(QFileDialog.DirectoryOnly)
        dialog.filesSelected.connect(
            lambda dirs: self.onDirsSelected(dialog, dirs))

        return dialog.exec_()

    def onDirsSelected(self, dialog, dirs):
        self.dialogLastDirSelected = dialog.directory()

        if isinstance(dirs, list):
            self.scheduleAddDirectory(dirs.pop())

    def statusAdded(self, name):
        self.statusSet(iAddedFile(name))

    def statusLoading(self, name):
        self.statusSet(iLoading(name))

    def statusSet(self, msg):
        self.ui.statusLabel.setText(msg)

    def onAddFilesClicked(self):
        dialog = QFileDialog(None)
        dialog.setFileMode(QFileDialog.ExistingFiles)

        dialog.filesSelected.connect(
            lambda files: self.onFilesSelected(dialog, files))
        dialog.exec_()

    def onFilesSelected(self, dialog, files):
        self.dialogLastDirSelected = dialog.directory()
        self.scheduleAddFiles(files, parent=self.displayPath)

    def scheduleAddFiles(self, path, parent=None):
        if self.busy:
            return

        if parent is None:
            parent = self.displayPath

        return self.app.task(self.addFiles, path, parent)

    def scheduleAddDirectory(self, path):
        if self.busy or self.inEncryptedFolder:
            return

        return self.app.task(self.addDirectory, path)

    def scheduleUnlink(self, item, cid):
        reply = questionBox('Unlink',
                            'Do you really want to unlink this item ?')
        if reply:
            ensure(self.unlinkFileFromHash(item, cid))

    def scheduleDelete(self, item, cid):
        reply = questionBox(
            'Delete',
            'Do you really want to delete: <b>{}</b>'.format(cid)
        )
        if reply:
            ensure(self.deleteFromCID(item, cid))

    def updateTree(self):
        self.app.task(self.listFiles, self.displayPath,
                      parentItem=self.displayedItem, maxdepth=1)

    @ipfsOp
    async def updateKeys(self, ipfsop):
        self.ipfsKeys = await ipfsop.keys()

    @ipfsOp
    async def makeDir(self, ipfsop, parent, path):
        await ipfsop.filesMkdir(os.path.join(parent, path))
        self.updateTree()

    @ipfsOp
    async def listFiles(self, ipfsop, path, parentItem, maxdepth=0,
                        autoexpand=False, timeout=20):
        if self.busy:
            return

        self.enableButtons(flag=False)
        self.status = self.statusBusy

        await self.listPathWithTimeout(ipfsop, path, parentItem,
                                       maxdepth=maxdepth,
                                       autoexpand=autoexpand,
                                       timeout=timeout)

        self.enableButtons()
        self.status = self.statusReady

    async def listPathWithTimeout(self, ipfsop, path, parentItem, **kw):
        timeout = kw.pop('timeout', 10)
        try:
            with async_timeout.timeout(timeout):
                await self.listPath(ipfsop, path, parentItem, **kw)
        except asyncio.TimeoutError:
            self.statusSet(iListingMFSTimeout(path))
        except aioipfs.APIError as err:
            self.statusSet(iIpfsError(err.message))
        except Exception:
            pass

    async def listPath(self, op, path, parentItem, depth=0, maxdepth=1,
                       autoexpand=False, timeout=10):
        if not parentItem or not parentItem.path:
            return

        self.statusSet(iListingMFSPath(path))

        listing = await op.filesList(path)
        if not listing:
            return

        parentItemSibling = self.model.sibling(parentItem.row(), 2,
                                               parentItem.index())
        parentItemHash = self.model.data(parentItemSibling)

        for entry in listing:
            await asyncio.sleep(0)

            found = False
            for child in parentItem.childrenItems():
                if child.entry['Name'] == entry['Name']:
                    if entry['Hash'] != child.entry['Hash']:
                        # The parent has a child item with the same
                        # filename but a different cid: this item
                        # was deleted and needs to be purged from the model
                        # to let the new entry show up
                        await modelhelpers.modelDeleteAsync(
                            self.model, child.entry['Hash'])
                        break
                if child.entry['Hash'] == entry['Hash']:
                    found = True
                    break

            if found:
                continue

            icon = None
            if entry['Type'] == 1:  # directory
                icon = self.iconFolder

            cidString = cidConvertBase32(entry['Hash'])
            nItemName = MFSNameItem(entry, entry['Name'], icon)
            nItemName.mimeFromDb(self.app.mimeDb)
            nItemName.setParentHash(parentItemHash)
            nItemSize = MFSItem(sizeFormat(entry['Size']))
            nItemCID = MFSItem(cidString)
            nItemCID.setToolTip(cidInfosMarkup(cidString))

            if not icon and nItemName.mimeTypeName:
                # If we have a better icon matching the file's type..

                mType = MIMEType(nItemName.mimeTypeName)
                mIcon = getIconFromMimeType(mType, defaultIcon='unknown')

                if mIcon:
                    nItemName.setIcon(mIcon)

            # Set its path in the MFS
            nItemName.path = os.path.join(parentItem.path, entry['Name'])

            nItem = [nItemName, nItemSize, nItemCID]

            parentItem.appendRow(nItem)

            if entry['Type'] == 1:  # directory
                if autoexpand is True:
                    self.ui.treeFiles.setExpanded(nItemName.index(), True)

                if maxdepth > depth:
                    # We used to await listPath() here but it sucks
                    # tremendously if you have a dead CID in the MFS which
                    # will make the ls timeout and hang the filemanager.
                    # Instead, use listPathWithTimeout() in another task. The
                    # FM status will be set to ready before potential
                    # subfolders are being listed in background tasks, which
                    # is fine

                    ensure(self.listPathWithTimeout(
                        op,
                        nItemName.path,
                        nItemName,
                        maxdepth=maxdepth, depth=depth + 1))

        if autoexpand is True:
            self.ui.treeFiles.expand(parentItem.index())

        self.model.refreshed.emit(parentItem)

    @ipfsOp
    async def deleteFromCID(self, ipfsop, item, cid):
        _dir = os.path.dirname(item.path)

        entry = await ipfsop.filesLookupHash(_dir, cid)

        if entry:
            # Delete the entry in the MFS directory
            await ipfsop.filesDelete(_dir,
                                     entry['Name'], recursive=True)

            # Purge the item's cid in the model
            # Purge its parent as well because the parent cid will change
            await modelhelpers.modelDeleteAsync(self.model, cid)

            if item.parentHash:
                await modelhelpers.modelDeleteAsync(
                    self.model, item.parentHash)

            ensure(ipfsop.purge(cid))
            log.debug('{0}: deleted'.format(cid))

            self.updateTree()
        else:
            log.debug('Did not find cid {}'.format(cid))

    @ipfsOp
    async def unlinkFileFromHash(self, op, item, cid):
        _dir = os.path.dirname(item.path)

        listing = await op.filesList(_dir)
        for entry in listing:
            await op.sleep()
            if entry['Hash'] == cid:
                await op.filesDelete(_dir, entry['Name'], recursive=True)
                await modelhelpers.modelDeleteAsync(self.model, cid)

    @ipfsOp
    async def addFilesSelfEncrypt(self, op, files, parent):
        self.enableButtons(flag=False)
        self.status = self.statusBusy

        async def onEntry(entry):
            self.statusAdded(entry['Name'])

        count = 0

        for file in files:
            await op.sleep()
            root = await op.addFileEncrypted(file)

            if root is None:
                self.statusSet(iFileImportError())
                continue

            base = os.path.basename(file)

            await self.linkEntry(op, root, parent, base)
            count += 1

        self.statusSet(iImportedCount(count))
        self.enableButtons()
        self.status = self.statusReady
        self.updateTree()
        return True

    @ipfsOp
    async def addFiles(self, op, files, parent):
        """ Add every file with an optional wrapper directory """

        if self.displayedItem is self.model.itemEncrypted and op.rsaAgent:
            return await self.addFilesSelfEncrypt(files, parent)

        wrapEnabled = self.app.settingsMgr.isTrue(
            CFG_SECTION_UI, CFG_KEY_WRAPSINGLEFILES)

        self.enableButtons(flag=False)
        self.status = self.statusBusy

        async def onEntry(entry):
            self.statusAdded(entry['Name'])

        count = 0
        for file in files:
            await op.sleep()

            root = await op.addPath(file, wrap=wrapEnabled,
                                    offline=self.offlineMode,
                                    dagformat=self.dagGenerationFormat,
                                    rawleaves=self.useRawLeaves,
                                    chunker=self.chunkingAlg,
                                    callback=onEntry)

            if root is None:
                self.statusSet(iFileImportError())
                continue

            base = os.path.basename(file)
            if wrapEnabled is True:
                base += '.dirw'

            await self.linkEntry(op, root, parent, base)
            count += 1

        self.statusSet(iImportedCount(count))
        self.enableButtons()
        self.status = self.statusReady
        self.updateTree()
        return True

    @ipfsOp
    async def addDirectory(self, op, path):
        wrapEnabled = self.app.settingsMgr.isTrue(
            CFG_SECTION_UI, CFG_KEY_WRAPDIRECTORIES)

        self.enableButtons(flag=False)
        self.status = self.statusBusy
        basename = os.path.basename(path)
        dirEntry = None

        async def onEntry(entry):
            self.statusAdded(entry['Name'])

        dirEntry = await op.addPath(path, callback=onEntry,
                                    hidden=True, recursive=True,
                                    offline=self.offlineMode,
                                    dagformat=self.dagGenerationFormat,
                                    rawleaves=self.useRawLeaves,
                                    chunker=self.chunkingAlg,
                                    wrap=wrapEnabled)

        if not dirEntry:
            # Nothing went through ?
            self.enableButtons()
            return False

        if wrapEnabled is True:
            basename += '.dirw'

        await self.linkEntry(op, dirEntry, self.displayPath, basename)

        self.enableButtons()
        self.status = self.statusReady
        self.updateTree()
        return True

    async def linkEntry(self, op, entry, dest, basename):
        for lIndex in range(0, 16):
            await op.sleep()
            if lIndex == 0:
                lNew = basename
            else:
                lNew = '{0}.{1}'.format(basename, lIndex)
            lookup = await op.filesLookup(dest, lNew)
            if not lookup:
                await op.sleep()
                linkS = await op.filesLink(entry, dest, name=lNew)
                if linkS:
                    return lNew


class FileManagerTab(GalacteekTab):
    def __init__(self, gWindow, fileManager=None):
        super(FileManagerTab, self).__init__(gWindow)

        self.fileManager = fileManager if fileManager else \
            FileManager(parent=self)
        self.fileManager.setupModel()

        self.addToLayout(self.fileManager)



import functools
import aioipfs
import asyncio
import random
from datetime import datetime

from PyQt5.QtWidgets import QAction
from PyQt5.QtWidgets import QToolTip
from PyQt5.QtWidgets import QApplication
from PyQt5.QtWidgets import QToolButton
from PyQt5.QtWidgets import QToolBar
from PyQt5.QtWidgets import QSizePolicy

from PyQt5.QtCore import QCoreApplication
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtCore import Qt
from PyQt5.QtCore import QSize
from PyQt5.QtCore import QBuffer
from PyQt5.QtCore import QPoint

from galacteek import ensure
from galacteek import log
from galacteek import logUser
from galacteek.ipfs.cidhelpers import IPFSPath
from galacteek.ipfs.cidhelpers import joinIpns
from galacteek.ipfs import ipfsOp
from galacteek.ipfs.stat import StatInfo
from galacteek.ipfs.dag import EvolvingDAG
from galacteek.ipfs.dag import DAGRewindException
from galacteek.dweb.render import ipfsRender
from galacteek.core import utcDatetimeIso
from galacteek.core.ipfsmarks import MultihashPyramid
from galacteek.core.profile import UserProfile
from galacteek.core.objects import SingleASig
from galacteek.crypto.qrcode import IPFSQrEncoder

from .widgets import PopupToolButton
from .widgets import URLDragAndDropProcessor
from .helpers import getMimeIcon
from .helpers import getIcon
from .helpers import getIconFromIpfs
from .helpers import runDialog
from .helpers import questionBox
from .helpers import getImageFromIpfs
from .helpers import inputTextLong
from .dialogs import AddMultihashPyramidDialog

from .i18n import iRemove
from .i18n import iHelp


def iCreateRawPyramid():
    return QCoreApplication.translate(
        'pyramidMaster',
        'Create pyramid (basic)')


def iCreateGallery():
    return QCoreApplication.translate(
        'pyramidMaster',
        'Create image gallery')


def iOpenLatestInPyramid():
    return QCoreApplication.translate(
        'pyramidMaster',
        'Open latest item in the pyramid')


def iEmptyPyramid():
    return QCoreApplication.translate(
        'pyramidMaster',
        'Pyramid is empty')


def iPopItemFromPyramid():
    return QCoreApplication.translate(
        'pyramidMaster',
        'Pop item off the pyramid')


def iRewindDAG():
    return QCoreApplication.translate(
        'pyramidMaster',
        'Rewind DAG')


def iRewindDAGToolTip():
    return QCoreApplication.translate(
        'pyramidMaster',
        "Rewinding the DAG cancels the latest "
        "operation/transformation in the DAG's history")


def iCopyIpnsAddress():
    return QCoreApplication.translate(
        'pyramidMaster',
        'Copy IPNS address to clipboard')


def iPyramidPublishCurrentClipboard():
    return QCoreApplication.translate(
        'pyramidMaster',
        'Add current clipboard item to the pyramid'
    )


def iPyramidGenerateQr():
    return QCoreApplication.translate(
        'pyramidMaster',
        "Generate pyramid's QR code"
    )


def iPyramidGenerateIndexQr():
    return QCoreApplication.translate(
        'pyramidMaster',
        "Generate index's QR code"
    )


def iGalleryBrowse():
    return QCoreApplication.translate(
        'pyramidMaster',
        "Browse image gallery"
    )


def iGalleryBrowseIpns():
    return QCoreApplication.translate(
        'pyramidMaster',
        "Browse image gallery (IPNS)"
    )


def iGalleryChangeTitle():
    return QCoreApplication.translate(
        'pyramidMaster',
        "Change image gallery's title"
    )


class MultihashPyramidsToolBar(QToolBar):
    moved = pyqtSignal(int)
    mfsInit = pyqtSignal(UserProfile)

    def __init__(self, parent):
        super().__init__(parent)

        self.app = QApplication.instance()
        self.app.marksLocal.pyramidConfigured.connect(self.onPyramidConfigured)
        self.app.marksLocal.pyramidNeedsPublish.connect(self.publishNeeded)
        self.app.marksLocal.pyramidChanged.connect(self.onPyramidChanged)
        self.app.marksLocal.pyramidEmpty.connect(self.onPyramidEmpty)

        self.setObjectName('toolbarPyramids')
        self.setToolTip('Hashmark pyramids toolbar')
        self.setContextMenuPolicy(Qt.PreventContextMenu)
        self.setFloatable(False)
        self.setMovable(False)
        self.setAcceptDrops(True)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.setOrientation(Qt.Vertical)

        pyrIcon = getIcon('pyramid-blue.png')
        self.pyramidsControlButton = PopupToolButton(
            icon=pyrIcon,
            mode=QToolButton.InstantPopup
        )
        self.pyramidsControlButton.setObjectName('pyramidsController')
        self.pyramidsControlButton.menu.addAction(
            getIcon('pyramid-aqua.png'),
            iCreateRawPyramid(), self.onAddPyramidRaw)
        self.pyramidsControlButton.menu.addSeparator()
        self.pyramidsControlButton.menu.addAction(
            getMimeIcon('image/x-generic'),
            iCreateGallery(), self.onAddGallery)

        self.pyramidsControlButton.menu.addSeparator()

        self.pyramidsControlButton.menu.addAction(
            pyrIcon, iHelp(), self.pyramidHelpMessage)
        self.addWidget(self.pyramidsControlButton)
        self.addSeparator()

        self.pyramids = {}
        self.setIconSize(QSize(32, 32))

    def contextMenuEvent(self, event):
        pass

    def setLargerIcons(self):
        current = self.iconSize()
        newSize = QSize(current.width() * 2, current.height() * 2)
        self.setIconSize(newSize)

    def removePyramidAsk(self, pyramidButton, action):
        reply = questionBox(
            iRemove(),
            'Remove pyramid <b>{pyr}</b> and its IPNS key ?'.format(
                pyr=pyramidButton.pyramid.name
            )
        )
        if reply is True:
            ensure(self.removePyramid(pyramidButton, action))

    async def removePyramid(self, pyramidButton, action):
        try:
            self.app.marksLocal.pyramidDrop(pyramidButton.pyramid.path)
            self.removeAction(action)

            await self.removeIpnsKey(pyramidButton.pyramid.ipnsKey)
            await pyramidButton.stop()
            await pyramidButton.cleanup()

            if pyramidButton.pyramid.path in self.pyramids:
                del self.pyramids[pyramidButton.pyramid.path]

            del pyramidButton
        except:
            log.debug('Pyramid remove failed')

    def pyramidHelpMessage(self):
        self.app.manuals.browseManualPage(
            'pyramids.html', fragment='multihash-pyramids')

    def onPyramidChanged(self, pyramidPath):
        if pyramidPath in self.pyramids:
            self.pyramids[pyramidPath].changed.emit()

    def onPyramidEmpty(self, pyramidPath):
        if pyramidPath in self.pyramids:
            self.pyramids[pyramidPath].emptyNow.emit()

    def onPyramidConfigured(self, pyramidPath):
        if pyramidPath in self.pyramids:
            # Already configured ?
            return

        pyramid = self.app.marksLocal.pyramidGet(pyramidPath)

        if pyramid is None:
            return

        if pyramid.type == MultihashPyramid.TYPE_STANDARD:
            button = MultihashPyramidToolButton(pyramid, parent=self)

        elif pyramid.type == MultihashPyramid.TYPE_GALLERY:
            button = GalleryPyramidController(pyramid, parent=self)
        else:
            # TODO
            return

        action = self.addWidget(button)
        button.deleteRequest.connect(functools.partial(
            self.removePyramidAsk, button, action))

        if pyramid.icon:
            ensure(self.fetchIcon(button, pyramid.icon))
        else:
            button.setIcon(getMimeIcon('unknown'))

        self.pyramids[pyramidPath] = button

    def publishNeeded(self, pyramidPath, mark):
        if pyramidPath in self.pyramids:
            pyramidMaster = self.pyramids[pyramidPath]
            ensure(pyramidMaster.needsPublish.send(mark, True))

    @ipfsOp
    async def fetchIcon(self, ipfsop, button, iconPath):
        icon = await getIconFromIpfs(ipfsop, iconPath)
        if icon:
            button.setIcon(icon)
        else:
            button.setIcon(getMimeIcon('unknown'))

    @ipfsOp
    async def removeIpnsKey(self, ipfsop, ipnsKeyId):
        """
        Remove the IPNS key associated with a pyramid (called on remove)
        """
        try:
            key = await ipfsop.keyFindById(ipnsKeyId)
        except aioipfs.APIError:
            # log here
            pass
        else:
            if key:
                log.info('Removing IPNS key: {}'.format(key))
                await ipfsop.keysRemove(key['Name'])

    def onAddPyramidRaw(self):
        runDialog(AddMultihashPyramidDialog, self.app.marksLocal,
                  MultihashPyramid.TYPE_STANDARD,
                  title='New multihash pyramid',
                  parent=self)

    def onAddGallery(self):
        runDialog(AddMultihashPyramidDialog, self.app.marksLocal,
                  MultihashPyramid.TYPE_GALLERY,
                  title='New image gallery',
                  parent=self)


class MultihashPyramidToolButton(PopupToolButton):
    deleteRequest = pyqtSignal()
    changed = pyqtSignal()
    emptyNow = pyqtSignal()

    def __init__(self, pyramid, icon=None, parent=None):
        super(MultihashPyramidToolButton, self).__init__(
            mode=QToolButton.InstantPopup, parent=parent)
        self.app = QApplication.instance()

        if icon:
            self.setIcon(icon)

        self.active = True
        self._publishInProgress = False
        self._pyramid = pyramid
        self._pyramidion = None
        self._publishedLast = None
        self._publishFailedCount = 0
        self._publishTimeout = 60 * 3
        self._pyrToolTip = None

        self.setAcceptDrops(True)
        self.setObjectName('pyramidMaster')

        self.clipboardItem = None
        self.app.clipTracker.currentItemChanged.connect(
            self.onClipboardItemChange)

        self.changed.connect(self.onPyrChange)
        self.emptyNow.connect(self.onPyrEmpty)

        self.needsPublish = SingleASig(self.pyramidNeedsPublish, self)

        self.ipfsObjectDropped.connect(self.onObjectDropped)
        self.clicked.connect(self.onOpenLatest)
        self.updateToolTip()

        self.openLatestAction = QAction(getIcon('pyramid-aqua.png'),
                                        iOpenLatestInPyramid(),
                                        self,
                                        triggered=self.onOpenLatest)
        self.openLatestAction.setEnabled(False)

        self.publishCurrentClipAction = QAction(
            getIcon('clipboard.png'),
            iPyramidPublishCurrentClipboard(),
            self,
            triggered=self.onPublishClipboardItem
        )
        self.publishCurrentClipAction.setEnabled(False)

        self.popItemAction = QAction(getIcon('pyramid-stack.png'),
                                     iPopItemFromPyramid(),
                                     self,
                                     triggered=self.onPopItem)
        self.popItemAction.setEnabled(False)

        self.copyIpnsAction = QAction(getIcon('clipboard.png'),
                                      iCopyIpnsAddress(),
                                      self,
                                      triggered=self.onCopyIpns)
        self.generateQrAction = QAction(getIcon('ipfs-qrcode.png'),
                                        iPyramidGenerateQr(),
                                        self,
                                        triggered=self.onGenerateQrCode)
        self.deleteAction = QAction(getIcon('cancel.png'),
                                    iRemove(),
                                    self,
                                    triggered=self.onDeletePyramid)

        self.createExtraActions()
        self.buildMenu()
        self.resetStyleSheet()
        ensure(self.initialize())
        self.watcherTask = self.app.task(self.publishWatcherTask)

    @property
    def pyramid(self):
        return self._pyramid

    @property
    def ipnsKeyPath(self):
        if self.pyramid.ipnsKey:
            return IPFSPath(joinIpns(self.pyramid.ipnsKey))

    @property
    def pyramidion(self):
        return self._pyramidion

    @pyramidion.setter
    def pyramidion(self, capStone):
        self._pyramidion = capStone

        self.openLatestAction.setEnabled(capStone is not None)
        self.popItemAction.setEnabled(capStone is not None)

        self.debug('Pyramidion changed: {top}'.format(
            top=capStone if capStone else 'empty'))

    @property
    def publishInProgress(self):
        return self._publishInProgress

    @property
    def publishedLast(self):
        return self._publishedLast

    @publishInProgress.setter
    def publishInProgress(self, value):
        if isinstance(value, bool):
            self._publishInProgress = value
            if value is True:
                self.setStyleSheet('''
                    QToolButton {
                        background-color: #B7CDC2;
                    }
                ''')
            else:
                self.app.loop.call_later(2, self.resetStyleSheet)

    @property
    def pyrToolTip(self):
        return self._pyrToolTip

    def debug(self, msg):
        logUser.debug('{pyramid}: {msg}'.format(
            pyramid=self.pyramid.path,
            msg=msg
        ))

    def info(self, msg):
        logUser.info('{pyramid}: {msg}'.format(
            pyramid=self.pyramid.path,
            msg=msg
        ))

    async def cleanup(self):
        pass

    async def stop(self):
        if self.watcherTask:
            try:
                self.watcherTask.cancel()
            except BaseException:
                self.debug('Could not cancel watcher task')

    async def initialize(self):
        mark = self.app.marksLocal.pyramidGetLatestHashmark(self.pyramid.path)
        if mark:
            self.pyramidion = mark

    def createExtraActions(self):
        pass

    def buildMenu(self):
        self.menu.addAction(self.openLatestAction)
        self.menu.addSeparator()
        self.menu.addAction(self.publishCurrentClipAction)
        self.menu.addSeparator()
        self.menu.addAction(self.popItemAction)
        self.menu.addAction(self.copyIpnsAction)
        self.menu.addAction(self.generateQrAction)
        self.menu.addSeparator()
        self.menu.addAction(self.deleteAction)

    def dragEnterEvent(self, event):
        URLDragAndDropProcessor.dragEnterEvent(self, event)

        if self.pyrToolTip:
            self.flashToolTip(self.pyrToolTip)
        else:
            self.flashToolTip('Pyramid: {path}'.format(path=self.pyramid.path))

        self.setStyleSheet('''
            QToolButton {
                background-color: #EB2121;
            }
        ''')

    def dropEvent(self, event):
        URLDragAndDropProcessor.dropEvent(self, event)
        self.resetStyleSheet()

    def dragLeaveEvent(self, event):
        self.resetStyleSheet()

    def resetStyleSheet(self):
        self.setStyleSheet('''
            QToolButton::hover {
                background-color: #4a9ea1;
            }

            QToolButton::pressed {
                background-color: #eec146;
            }
        ''')

    def flashToolTip(self, message):
        QToolTip.showText(self.mapToGlobal(QPoint(0, 0)), message)

    def updateToolTip(self):
        self._pyrToolTip = '''
            <p>
                <img width='64' height='64'
                    src=':/share/icons/pyramid-hierarchy.png'/>
            </p>
            <p>
                Multihash pyramid: <b>{path}</b>
                ({itemscount} item(s) in the stack)
            </p>
            <p>Description: {descr}</p>
            <p>IPNS key: <b>{ipns}</b></p>

            <p>
                <img width='16' height='16'
                    src=':/share/icons/pyramid-stack.png'/>
                Latest (pyramidion): <b>{latest}</b>
            </p>
        '''.format(
            path=self.pyramid.path,
            descr=self.pyramid.description,
            ipns=self.pyramid.ipnsKey,
            itemscount=self.pyramid.marksCount,
            latest=self.pyramid.latest if self.pyramid.latest else
            iEmptyPyramid()
        )
        self.setToolTip(self.pyrToolTip)

    def onDeletePyramid(self):
        self.deleteRequest.emit()

    def onPyrChange(self):
        self.updateToolTip()

    def onPyrEmpty(self):
        self.pyramidion = None
        ensure(self.publishEmptyObject())

    def onGenerateQrCode(self):
        if self.ipnsKeyPath:
            qrName = 'ipfsqr.{}.png'.format(self.pyramid.ipnsKey)
            ensure(self.generateQrCode(qrName, self.ipnsKeyPath))

    def onObjectDropped(self, ipfsPath):
        self.app.marksLocal.pyramidAdd(self.pyramid.path, str(ipfsPath))
        self.updateToolTip()
        self.app.systemTrayMessage(
            'Pyramids',
            'Pyramid {pyr}: registered new hashmark: {path}'.format(
                pyr=self.pyramid.path, path=str(ipfsPath)))

    def sysTray(self, message):
        self.app.systemTrayMessage(
            'Pyramids',
            'Pyramid {path}: {msg}'.format(
                path=self.pyramid.path, msg=message))

    def onPopItem(self):
        res = self.app.marksLocal.pyramidPop(self.pyramid.path)
        if res is True:
            self.sysTray('Item popped')

            if self.pyramid.empty:
                self.sysTray('Pyramid is empty now!')

    def onOpenLatest(self):
        """ Open latest object """
        if not isinstance(self.pyramid.latest, str) or not self.pyramidion:
            return

        objPath = IPFSPath(self.pyramid.latest, autoCidConv=True)
        if objPath.valid:
            ensure(self.app.resourceOpener.open(objPath))

    def onClipboardItemChange(self, clipItem):
        self.clipboardItem = clipItem
        self.publishCurrentClipAction.setEnabled(True)

    def onPublishClipboardItem(self):
        if self.clipboardItem:
            reply = questionBox(
                'Publish',
                'Publish clipboard item (<b>{0}</b>) to this pyramid ?'.format(
                    str(self.clipboardItem.path))
            )
            if reply is True:
                self.app.marksLocal.pyramidAdd(
                    self.pyramid.path, str(self.clipboardItem.path)
                )

    def onCopyIpns(self):
        self.app.setClipboardText(joinIpns(self.pyramid.ipnsKey))

    @ipfsOp
    async def generateQrCode(self, ipfsop, qrName, *paths):
        encoder = IPFSQrEncoder()

        for path in paths:
            encoder.add(str(path))

        imgPath = self.app.tempDir.filePath(qrName)

        try:
            image = await encoder.encodeAll(loop=self.app.loop,
                                            executor=self.app.executor)
            image.save(imgPath)
        except:
            logUser.info('QR: encoding error ..')
            return
        else:
            logUser.info('QR: encoding successfull!')

        if ipfsop.ctx.currentProfile:
            ipfsop.ctx.currentProfile.qrImageEncoded.emit(False, imgPath)

    @ipfsOp
    async def publishObject(self, ipfsop, objPath):
        return await ipfsop.publish(
            objPath,
            key=self.pyramid.ipnsKey,
            allow_offline=self.pyramid.ipnsAllowOffline,
            lifetime=self.pyramid.ipnsLifetime,
            timeout=self._publishTimeout
        )

    @ipfsOp
    async def publishEmptyObject(self, ipfsop):
        try:
            if ipfsop.ctx.softIdent:
                await self.publishObject(
                    ipfsop.ctx.softIdent['Hash']
                )
        except:
            pass

    @ipfsOp
    async def publish(self, ipfsop, latestMark, notify=True):
        try:
            if latestMark.path is None:
                return False

            ipfsPath = IPFSPath(latestMark.path, autoCidConv=True)
            if not ipfsPath.valid:
                self.debug('Invalid path! Cannot publish')
                return False

            self.info('publishing mark {mark} (obj: {obj}) to {ipns}'.format(
                mark=latestMark.path,
                obj=ipfsPath.objPath,
                ipns=self.pyramid.ipnsKey
            ))

            self.publishInProgress = True

            result = await self.publishObject(
                ipfsPath.objPath
            )
        except aioipfs.APIError as err:
            self.publishInProgress = False
            self.debug('Publish error: {msg}'.format(msg=err.message))
            return False
        except Exception:
            self.publishInProgress = False
            self.debug('Unknown exception while publishing')
            return False
        else:
            self.publishInProgress = False

            if result:
                # Publish successfull
                self._publishedLast = datetime.now()

                if notify is True:
                    self.sysTray('Pyramid was published!')

                return True
            else:
                self._publishFailedCount += 1
                self.info('Publish failed: ({count} error(s))'.format(
                    count=self._publishFailedCount))
                return False

    async def pyramidNeedsPublish(self, mark, notify):
        """
        We need to publish! Unless there's already a publish
        in progress, ensure an update right away
        """

        self.pyramidion = mark

        if self.publishInProgress is False:
            await self.publish(self.pyramidion, notify)

    async def publishWatcherTask(self):
        # Depending on the lifetime of the records we publish, decide upon
        # a maximum time for us not to trigger a republish
        #
        # We sorta anticipate the record expiring
        # We be nice to the network

        if self.pyramid.ipnsLifetime == '96h':
            unpublishedMax = 84 * 3600
        elif self.pyramid.ipnsLifetime == '48h':
            unpublishedMax = 42 * 3600
        elif self.pyramid.ipnsLifetime == '24h':
            unpublishedMax = 18 * 3600
        elif self.pyramid.ipnsLifetime == '12h':
            unpublishedMax = 10 * 3600
        else:
            unpublishedMax = None

        while self.active:
            if self.pyramidion:
                if self.publishedLast is None:
                    # Publish on startup after random delay
                    rand = random.Random()
                    delay = rand.randint(5, 30)

                    self.app.loop.call_later(
                        delay, ensure, self.needsPublish.send(self.pyramidion,
                                                              False))

                if isinstance(self.publishedLast, datetime):
                    delta = datetime.now() - self.publishedLast

                    if unpublishedMax and delta.seconds > unpublishedMax:
                        await self.needsPublish.send(self.pyramidion, True)

            await asyncio.sleep(3600)


class EDAGBuildingPyramidController(MultihashPyramidToolButton):
    """
    A type of pyramid that works on a EDAG, like the gallery generator
    """

    edagClass = EvolvingDAG

    def onObjectDropped(self, ipfsPath):
        pass

    def buildMenu(self):
        pass

    def createExtraActions(self):
        self.rewindDagAction = QAction(getIcon('ipld.png'),
                                       iRewindDAG(),
                                       self,
                                       triggered=self.onRewindEDAG)
        self.rewindDagAction.setToolTip(iRewindDAGToolTip())

    def onRewindEDAG(self):
        ensure(self.rewindEDAG())

    async def rewindEDAG(self):
        """
        Rewind the EDAG one iteration.

        Pop a hashmark off the pyramid, and rewind the EDAG.

        TODO: improve the EDAG rewind API so that we can know
        which object we're moving away from
        """
        try:
            self.app.marksLocal.pyramidPop(self.pyramid.path)
            await self.edag.rewind()
        except DAGRewindException:
            self.info('Cannot rewind DAG (no DAG history)')
        else:
            self.info('DAG rewind successfull')

    @ipfsOp
    async def cleanup(self, ipfsop):
        """
        Remove the EDAG's metadata file.
        We should probably also unpin the latest EDAG's object
        """
        await super().cleanup()
        await ipfsop.filesRm(self.edagMetaPath)

    async def stop(self):
        await super().stop()

    @ipfsOp
    async def initialize(self, ipfsop):
        profile = ipfsop.ctx.currentProfile
        self.edagMetaPath = profile.edagPyramidMetadataPath(self.pyramid.uuid)
        await self.initDag()

    @ipfsOp
    async def initDag(self, ipfsop):
        self.info('Loading EDAG from MFS metadata: {}'.format(
            self.edagMetaPath))

        self.edag = self.edagClass(self.edagMetaPath)
        ensure(self.edag.load())
        await self.edag.loaded

        await self.initDagExtra(self.edag)

        self.edag.dagCidChanged.connect(self.onPyramidDagCidChanged)

        mark = self.app.marksLocal.pyramidGetLatestHashmark(self.pyramid.path)
        if mark:
            self.pyramidion = mark

    @ipfsOp
    async def initDagExtra(self, ipfsop, edag):
        pass

    def onPyramidDagCidChanged(self, cidStr):
        self.info("Pyramid's DAG moving to CID: {}".format(cidStr))

        path = IPFSPath(cidStr)

        if path.valid:
            self.app.marksLocal.pyramidAdd(self.pyramid.path, str(path))


class GalleryDAG(EvolvingDAG):
    def updateDagSchema(self, root):
        changed = False
        if 'images' not in root:
            root['images'] = []
            changed = True

        if 'metadata' not in root:
            root['metadata'] = {
                'author': None,
                'title': 'Image gallery',
                'creationdate': utcDatetimeIso()
            }
            changed = True

        return changed


class GalleryPyramidController(EDAGBuildingPyramidController):
    edagClass = GalleryDAG

    @property
    def indexPath(self):
        return IPFSPath(self.pyramid.latest, autoCidConv=True).child(
            'index.html')

    @property
    def indexIpnsPath(self):
        return self.ipnsKeyPath.child('index.html')

    def onObjectDropped(self, ipfsPath):
        ensure(self.analyzeImageObject(ipfsPath))

    def createExtraActions(self):
        super().createExtraActions()
        self.browseIpnsAction = QAction(getIcon('pyramid-aqua.png'),
                                        iGalleryBrowseIpns(),
                                        self,
                                        triggered=self.onBrowseGalleryIpns)
        self.browseDirectAction = QAction(getIcon('pyramid-aqua.png'),
                                          iGalleryBrowse(),
                                          self,
                                          triggered=self.onBrowseGallery)
        self.generateIndexQrAction = QAction(
            getIcon('ipfs-qrcode.png'),
            iPyramidGenerateIndexQr(),
            self,
            triggered=self.onGenerateIndexQrCode)
        self.changeTitleAction = QAction(getIcon('pyramid-aqua.png'),
                                         iGalleryChangeTitle(),
                                         self,
                                         triggered=self.onChangeTitle)

    def buildMenu(self):
        self.menu.addAction(self.browseIpnsAction)
        self.menu.addAction(self.browseDirectAction)
        self.menu.addSeparator()
        self.menu.addAction(self.copyIpnsAction)
        self.menu.addSeparator()
        self.menu.addAction(self.changeTitleAction)
        self.menu.addSeparator()
        self.menu.addAction(self.generateIndexQrAction)
        self.menu.addAction(self.generateQrAction)
        self.menu.addSeparator()
        self.menu.addAction(self.rewindDagAction)
        self.menu.addSeparator()
        self.menu.addAction(self.deleteAction)
        self.menu.addSeparator()
        self.menu.addAction(
            getIcon('pyramid-blue.png'), iHelp(), self.galleryHelpMessage)
        self.menu.setEnabled(False)
        self.menu.setToolTipsVisible(True)

    def onChangeTitle(self):
        curTitle = self.edag.root['metadata']['title']

        title = inputTextLong(
            label='Gallery title',
            text=curTitle
        )

        if title and title != curTitle:
            ensure(self.changeGalleryTitle(title))

    @ipfsOp
    async def changeGalleryTitle(self, ipfsop, title):
        async with self.edag as edag:
            edag.root['metadata']['title'] = title

    def onGenerateIndexQrCode(self):
        qrName = 'ipfsqr.pyrindex.{}.png'.format(self.pyramid.name)
        ensure(self.generateQrCode(qrName, self.indexIpnsPath))

    def onBrowseGalleryIpns(self):
        objPath = self.ipnsKeyPath.child('index.html')
        ensure(self.app.resourceOpener.open(objPath))

    def onBrowseGallery(self):
        if self.pyramid.latest:
            objPath = IPFSPath(self.pyramid.latest, autoCidConv=True).child(
                'index.html')
            ensure(self.app.resourceOpener.open(objPath))

    @ipfsOp
    async def initDagExtra(self, ipfsop, edag):
        if 'index.html' not in edag.root:
            entry = await ipfsRender(self.app.jinjaEnv,
                                     'imggallery/gallery.html')
            edag.root['index.html'] = self.edag.mkLink(entry)
            edag.changed.emit()

        self.menu.setEnabled(True)

    @ipfsOp
    async def analyzeImageObject(self, ipfsop, ipfsPath):
        mimeType, stat = await self.app.rscAnalyzer(ipfsPath)

        if not mimeType or not mimeType.isImage:
            self.app.messageDisplayRequest.emit(
                'This object is not an image', 'Wrong object type')
            return

        statInfo = StatInfo(stat)
        image = await getImageFromIpfs(ipfsPath)

        if not image:
            # TODO
            return

        await ipfsop.sleep()

        imgName = ipfsPath.basename if ipfsPath.subPath else None
        tEntry = None
        maxSize = QSize(128, 128)

        # Create a thumbnail
        if image.size().width() > maxSize.width() or \
                image.size().height() > maxSize.height():

            thumbnail = await self.app.loop.run_in_executor(
                self.app.executor,
                image.scaled, maxSize, Qt.KeepAspectRatio)

            buff = QBuffer(parent=self)
            buff.open(QBuffer.ReadWrite)
            thumbnail.save(buff, 'PNG')
            buff.close()

            await ipfsop.sleep()

            data = bytes(buff.buffer())

            if len(data) > 0:
                tEntry = await ipfsop.addBytes(data)

        self.edag.root['images'].append({
            'comment': None,
            'name': imgName,
            'date': utcDatetimeIso(),
            'thumbnail': self.edag.mkLink(tEntry) if tEntry else None,
            'raw': self.edag.mkLink(statInfo.cid),
            'imginfo': {
                'width': image.width(),
                'height': image.height(),
                'depth': image.depth(),
                'bitplanes': image.bitPlaneCount(),
            }
        })

        entry = await ipfsRender(self.app.jinjaEnv,
                                 'imggallery/gallery.html')
        self.edag.root['index.html'] = self.edag.mkLink(entry)
        self.edag.changed.emit()

    def updateToolTip(self):
        self._pyrToolTip = '''
            <p>
                <img width='64' height='64'
                    src=':/share/icons/mimetypes/image-x-generic.png'/>
            </p>
            <p>
                Image gallery: <b>{path}</b>
                ({itemscount} item(s) in the stack)
            </p>

            <p><b>Drag-and-drop images here</b>
            to add an image to the gallery</p>

            <p>Description: {descr}</p>
            <p>IPNS key: <b>{ipns}</b></p>

            <p>
                <img width='16' height='16'
                    src=':/share/icons/pyramid-stack.png'/>
                Latest (pyramidion): <b>{latest}</b>
            </p>
        '''.format(
            path=self.pyramid.path,
            descr=self.pyramid.description,
            ipns=self.pyramid.ipnsKey,
            itemscount=self.pyramid.marksCount,
            latest=self.pyramid.latest if self.pyramid.latest else
            iEmptyPyramid()
        )
        self.setToolTip(self.pyrToolTip)

    def galleryHelpMessage(self):
        self.app.manuals.browseManualPage(
            'pyramids.html', fragment='image-gallery')



from PyQt5.QtCore import QCoreApplication


def qtr(ctx, msg):
    return QCoreApplication.translate(ctx, msg)


# CID

def iCID():
    return QCoreApplication.translate('GalacteekWindow', 'CID')


def iCIDv0():
    return QCoreApplication.translate('GalacteekWindow', 'CIDv0')


def iCIDv1():
    return QCoreApplication.translate('GalacteekWindow', 'CIDv1')


def iP2PKey():
    return QCoreApplication.translate('GalacteekWindow', 'P2P key')


def iUnixFSNode():
    return QCoreApplication.translate('GalacteekWindow', 'UnixFS node')


# Main window messages


def iNoStatus():
    return QCoreApplication.translate('GalacteekWindow', 'No status')


def iGeneralError(msg):
    return QCoreApplication.translate('GalacteekWindow',
                                      'General error: {0}').format(msg)


def iIpfsError(msg):
    return QCoreApplication.translate('GalacteekWindow',
                                      'IPFS error: {0}').format(msg)


def iErrNoCx():
    return QCoreApplication.translate(
        'GalacteekWindow',
        'No connection available')


def iCxButNoPeers(id, agent):
    return QCoreApplication.translate(
        'GalacteekWindow',
        'IPFS ({1}): not connected to any peers').format(
        id, agent)


def iConnectStatus(id, agent, peerscount):
    return QCoreApplication.translate(
        'GalacteekWindow',
        'IPFS ({1}): connected to {2} peer(s)').format(
        id, agent, peerscount)


def iItemsInPinningQueue(itemsCount):
    return QCoreApplication.translate(
        'GalacteekWindow',
        'Items queued for pinning: {}'.format(itemsCount))


def iBrowse():
    return QCoreApplication.translate('GalacteekWindow', 'Browse')


def iBrowseAutoPin():
    return QCoreApplication.translate('GalacteekWindow', 'Browse (auto-pin)')


def iHelp():
    return QCoreApplication.translate('GalacteekWindow', 'Help')


def iUnknown():
    return QCoreApplication.translate('GalacteekWindow', 'Unknown')


def iUnknownAgent():
    return QCoreApplication.translate('GalacteekWindow', 'Unknown agent')


def iMinimized():
    return QCoreApplication.translate(
        'GalacteekWindow',
        'galacteek was minimized to the system tray')


# Headers used in the various tree widgets
def iPath():
    return QCoreApplication.translate('IPFSTreeView', 'Path')


def iCidOrPath():
    return QCoreApplication.translate('IPFSTreeView', 'CID or path')


def iFileName():
    return QCoreApplication.translate('IPFSTreeView', 'Name')


def iFileSize():
    return QCoreApplication.translate('IPFSTreeView', 'Size')


def iFileHash():
    return QCoreApplication.translate('IPFSTreeView', 'Hash')


def iMultihash():
    return QCoreApplication.translate('IPFSTreeView', 'Multihash')


def iMimeType():
    return QCoreApplication.translate('IPFSTreeView', 'MIME type')


def iPinSuccess(path):
    return QCoreApplication.translate(
        'GalacteekWindow',
        '{0} was pinned successfully').format(path)


def iPinError(path, errmsg):
    return QCoreApplication.translate(
        'GalacteekWindow',
        'Error pinning {0}: {1}').format(path, errmsg)


def iManual():
    return QCoreApplication.translate('GalacteekWindow', 'Manual')


def iLinkToMfsFolder():
    return QCoreApplication.translate(
        'GalacteekWindow', 'Link to folder (MFS)')


def iHashmark():
    return QCoreApplication.translate('GalacteekWindow', 'Hashmark')


def iHashmarkThisPage():
    return QCoreApplication.translate('GalacteekWindow', 'Hashmark this page')


def iHashmarkRootObject():
    return QCoreApplication.translate(
        'GalacteekWindow', 'Hashmark root object')


def iHashmarks():
    return QCoreApplication.translate('GalacteekWindow', 'Hashmarks')


def iLocalHashmarks():
    return QCoreApplication.translate('GalacteekWindow', 'Local hashmarks')


def iSharedHashmarks():
    return QCoreApplication.translate('GalacteekWindow', 'Shared hashmarks')


def iSearchHashmarks():
    return QCoreApplication.translate('GalacteekWindow', 'Search hashmarks')


def iSearchUseShiftReturn():
    return QCoreApplication.translate(
        'GalacteekWindow',
        '<p>Press <b>Shift + Return</b> to run a search</p>')


def iSearchHashmarksAllAcross():
    return QCoreApplication.translate(
        'GalacteekWindow',
        '''
        <div>
            <p>Search the hashmarks database</p>

            <p>Press <b>Shift + Return</b> to validate your search</p>
        </div>
        ''')


def iHashmarkInfoToolTip(path, title, description, dateCreated):
    return QCoreApplication.translate(
        'GalacteekWindow',
        '''
            <img src=':/share/icons/hashmarks.png' width='16' height='16' />
            <p style='font: Courier 12pt'>{0}</p>

            <p>Title: <b>{1}</b></p>
            <p>Description: <b>{2}</b></p>

            <p>Creation date: <b>{3}</b></p>
        ''').format(path,
                    title if title else iNoTitle(),
                    description if description else iNoDescription(),
                    dateCreated
                    )


def iHashmarksLibraryCountAvailable(count):
    return QCoreApplication.translate(
        'GalacteekWindow',
        'Hashmarks library: {0} hashmarks available'
    ).format(count)


def iHashmarksLibrary():
    return QCoreApplication.translate('GalacteekWindow',
                                      '''
            <img src=':/share/icons/hashmarks-library.png'
                width='32' height='32' />

            <p>Hashmarks library</p>
        ''')


def iFileManager():
    return QCoreApplication.translate('GalacteekWindow', 'File Manager')


def iTextEditor():
    return QCoreApplication.translate('GalacteekWindow', 'Editor')


def iHashmarksManager():
    return QCoreApplication.translate('GalacteekWindow', 'Hashmarks manager')


def iKeys():
    return QCoreApplication.translate('GalacteekWindow', 'IPFS Keys')


def iSettings():
    return QCoreApplication.translate('GalacteekWindow', 'Settings')


def iPinningStatus():
    return QCoreApplication.translate('GalacteekWindow', 'Pinning status')


def iEventLog():
    return QCoreApplication.translate('GalacteekWindow', 'Event Log')


def iPeers():
    return QCoreApplication.translate('GalacteekWindow', 'Peers')


def iIpfsSearchText(text):
    return QCoreApplication.translate('GalacteekWindow',
                                      'Search: {0}').format(text)


def iIpfsSearch():
    return QCoreApplication.translate('GalacteekWindow',
                                      'Search')


def iNewProfile():
    return QCoreApplication.translate('GalacteekWindow', 'New Profile')


def iSwitchedProfile():
    return QCoreApplication.translate('GalacteekWindow',
                                      'Successfully switched profile')


def iDagViewer():
    return QCoreApplication.translate('GalacteekWindow', 'DAG viewer')


def iDagView():
    return QCoreApplication.translate('GalacteekWindow', 'DAG view')


def iParentDagView():
    return QCoreApplication.translate(
        'GalacteekWindow', 'DAG view (parent node)')


def iLocalHashmarksCount(count):
    return QCoreApplication.translate(
        'GalacteekWindow',
        'Local hashmarks: {0} hashmarks available'
    ).format(count)


def iOpen():
    return QCoreApplication.translate('GalacteekWindow', 'Open')


def iDownload():
    return QCoreApplication.translate('GalacteekWindow', 'Download')


def iCancel():
    return QCoreApplication.translate('GalacteekWindow', 'Cancel')


def iMediaPlayer():
    return QCoreApplication.translate('GalacteekWindow', 'Media Player')


def iChat():
    return QCoreApplication.translate('GalacteekWindow', 'Chat')


def iLangEnglish():
    return QCoreApplication.translate('Galacteek', 'English')


def iLangFrench():
    return QCoreApplication.translate('Galacteek', 'French')


def iYes():
    return QCoreApplication.translate('Galacteek', 'yes')


def iNo():
    return QCoreApplication.translate('Galacteek', 'no')


def iDelete():
    return QCoreApplication.translate('Galacteek', 'Delete')


def iRemove():
    return QCoreApplication.translate('Galacteek', 'Remove')


def iInvalidInput():
    return QCoreApplication.translate('GalacteekWindow', 'Invalid input')


def iInvalidCID():
    return QCoreApplication.translate('GalacteekWindow', 'Invalid CID')


def iCannotResolve(objPath):
    return QCoreApplication.translate(
        'Galacteek',
        'Cannot resolve object: <b>{}</b>').format(objPath)


def iKey():
    return QCoreApplication.translate('Galacteek', 'Key')


def iValue():
    return QCoreApplication.translate('Galacteek', 'Value')


def iMerkleLink():
    return QCoreApplication.translate('Galacteek', 'Merkle link')


def iIpfsInfos():
    return QCoreApplication.translate('GalacteekWindow', 'IPFS informations')


def iIpfsQrCodes():
    return QCoreApplication.translate('GalacteekWindow', 'IPFS QR codes')


def iIpfsQrEncode():
    return QCoreApplication.translate('GalacteekWindow', 'QR encoding')


def iTitle():
    return QCoreApplication.translate('GalacteekWindow', 'Title')


def iNoTitle():
    return QCoreApplication.translate('GalacteekWindow', 'No title')


def iNoDescription():
    return QCoreApplication.translate('GalacteekWindow', 'No description')


def iNoTitleProvided():
    return QCoreApplication.translate('GalacteekWindow',
                                      'Please specify a title')


def iFinished():
    return QCoreApplication.translate('GalacteekWindow', 'Finished')


def iPinned():
    return QCoreApplication.translate('GalacteekWindow', 'Pinned')


def iPinning():
    return QCoreApplication.translate('GalacteekWindow', 'Pinning')


def iPin():
    return QCoreApplication.translate('GalacteekWindow', 'Pin')


def iPinSingle():
    return QCoreApplication.translate('GalacteekWindow', 'Pin (single)')


def iPinRecursive():
    return QCoreApplication.translate('GalacteekWindow', 'Pin (recursive)')


def iPinRecursiveParent():
    return QCoreApplication.translate('GalacteekWindow',
                                      'Pin parent (recursive)')


def iPinPageLinks():
    return QCoreApplication.translate('GalacteekWindow', "Pin page's links")


def iBatchPin():
    return QCoreApplication.translate('GalacteekWindow', 'Batch pin')


def iDoNotPin():
    return QCoreApplication.translate('GalacteekWindow', 'Do not pin')


def iGlobalAutoPinning():
    return QCoreApplication.translate('GalacteekWindow',
                                      'Global automatic pinning')


def iZoomIn():
    return QCoreApplication.translate('GalacteekWindow', 'Zoom in')


def iZoomOut():
    return QCoreApplication.translate('GalacteekWindow', 'Zoom out')


def iQuit():
    return QCoreApplication.translate('GalacteekWindow', 'Quit')


def iRestart():
    return QCoreApplication.translate('GalacteekWindow', 'Restart')


def iClearHistory():
    return QCoreApplication.translate('GalacteekWindow', 'Clear history')


def iAtomFeeds():
    return QCoreApplication.translate('GalacteekWindow', 'Atom feeds')


# IPFS daemon ui messages

def iFsRepoMigrateNotFound():
    return QCoreApplication.translate(
        'Galacteek',
        'Warning: could not find IPFS repository migration tool')


def iGoIpfsNotFound():
    return QCoreApplication.translate(
        'Galacteek',
        'Error: Could not find go-ipfs on your system')


def iGoIpfsTooOld():
    return QCoreApplication.translate(
        'Galacteek',
        'Error: go-ipfs version found on your system is too old')


def iGoIpfsFetchAsk():
    return QCoreApplication.translate(
        'Galacteek',
        'go-ipfs was not found on your system: download '
        'binary from IPFS distributions website (https://dist.ipfs.io) ?')


def iGoIpfsFetchTimeout():
    return QCoreApplication.translate(
        'Galacteek',
        'Timeout while fetching go-ipfs distribution')


def iGoIpfsFetchSuccess():
    return QCoreApplication.translate(
        'Galacteek',
        'go-ipfs was installed successfully')


def iGoIpfsFetchError():
    return QCoreApplication.translate(
        'Galacteek',
        'Error while fetching go-ipfs distribution')


def iNewReleaseAvailable():
    return QCoreApplication.translate(
        'Galacteek',
        'New release available: upgrade with pip install -U galacteek')


# MFS

def iMusic():
    return QCoreApplication.translate('FileManagerForm', 'Music')


def iPictures():
    return QCoreApplication.translate('FileManagerForm', 'Pictures')


def iImages():
    return QCoreApplication.translate('FileManagerForm', 'Images')


def iTemporaryFiles():
    return QCoreApplication.translate('FileManagerForm', 'Temporary')


def iEncryptedFiles():
    return QCoreApplication.translate('FileManagerForm', 'Encrypted files')


def iVideos():
    return QCoreApplication.translate('FileManagerForm', 'Videos')


def iHome():
    return QCoreApplication.translate('FileManagerForm', 'Home')


def iCode():
    return QCoreApplication.translate('FileManagerForm', 'Code')


def iDocuments():
    return QCoreApplication.translate('FileManagerForm', 'Documents')


def iWebPages():
    return QCoreApplication.translate('FileManagerForm', 'Web Pages')


def iDWebApps():
    return QCoreApplication.translate('FileManagerForm', 'Dapps')


def iQrCodes():
    return QCoreApplication.translate('FileManagerForm', 'QR codes')



import asyncio
import time

from galacteek import log
from galacteek.core.ipfsmarks import *  # noqa
from galacteek.ipfs.ipfsops import *  # noqa
from galacteek.ipfs.wrappers import ipfsOp
from galacteek.ipfs import crawl


class FeedFollower(object):
    """
    Follows an IPNS hash reference, in the sense that it will periodically
    resolve the hash and register the resulting object path if it's new,
    adding it as a mark inside the feed object
    """

    def __init__(self, app, marks):
        self.marks = marks
        self.app = app

    async def process(self):
        try:
            await self.processIpnsFeeds()
        except asyncio.CancelledError:
            pass
        except asyncio.TimeoutError:
            pass
        except BaseException:
            log.debug('IPNS follower: unknown error ocurred')

    @ipfsOp
    async def processIpnsFeeds(self, op):
        while True:
            feeds = self.marks.getFeeds()

            for ipnsp, feed in feeds:
                now = int(time.time())

                resolvedLast = feed.get('resolvedlast', None)
                resolveEvery = feed.get('resolveevery', 3600)
                autoPin = feed.get('autopin', False)
                feedMarks = self.marks.getFeedMarks(ipnsp)

                if resolvedLast and resolvedLast > (now - resolveEvery):
                    log.debug('{0} was resolved recently'.format(ipnsp))
                    continue

                resolved = await op.nameResolve(ipnsp, timeout=15,
                                                recursive=True)
                if not resolved:
                    log.debug('Could not resolve {0}'.format(ipnsp))
                    continue

                resolvedPath = resolved.get('Path', None)
                if not resolvedPath:
                    log.debug(
                        'Could not resolve {0}: invalid path'.format(ipnsp))
                    continue

                feed['resolvedlast'] = now

                if resolvedPath in feedMarks.keys():
                    # already registered
                    log.debug('Not registering {0}'.format(resolvedPath))
                    continue

                # Register the mark
                objStats = {}
                stat = await op.objStat(resolvedPath, timeout=10)
                if stat:
                    objStats = stat

                title = await crawl.getTitle(op.client, resolvedPath)

                mark = IPFSHashMark.make(
                    resolvedPath, title=title,
                    datasize=objStats.get('DataSize', None),
                    cumulativesize=objStats.get('CumulativeSize', None),
                    numlinks=objStats.get('NumLinks', None),
                    share=feed.get('share', False)
                )

                self.marks.feedAddMark(ipnsp, mark)

                if autoPin:
                    log.debug('Feed follower, autopinning {}'.format(
                        resolvedPath))
                    await self.app.ipfsCtx.pinner.queue(resolvedPath,
                                                        True, None)

            await asyncio.sleep(120)
